import os
import copy
import glob
import json
import shutil
import asyncio
import binascii
import itertools
import contextlib
import collections

import synapse.exc as s_exc
import synapse.cortex as s_cortex
import synapse.common as s_common

import synapse.tests.utils as s_t_utils

import synapse.lib.cell as s_cell
import synapse.lib.msgpack as s_msgpack
import synapse.lib.version as s_version
import synapse.lib.lmdbslab as s_lmdbslab
import synapse.lib.modelrev as s_modelrev
import synapse.lib.stormsvc as s_stormsvc

import synapse.tools.migrate_200 as s_migr

REGR_VER = '0.1.56-migr'
REGR_VER_OLD = '0.1.51-migr'

# Nodes that are expected to be unmigratable
NOMIGR_NDEF = [
    ["migr:test", 22],
    ["test:int", 10],
]

# Error log to be generated by migr:test
# .created val removed since this will update when the regression repo is updated
MIGR_ERR = {
    'migrop': 'nodes',
    'logtyp': 'error',
    'key': b'\xc8ak\x08\xb8\x8fJ\xcd\xad/\xc4F\x08\x7f@\xc9k\xf8\xc4\xca\x807|\xacK\xe5\xff<\x88+7\xef',
    'val': {
        'mesg': "Unable to determine stortype for migr:test: 'NoneType' object has no attribute 'type'",
        'node': (
            b'\xc8ak\x08\xb8\x8fJ\xcd\xad/\xc4F\x08\x7f@\xc9k\xf8\xc4\xca\x807|\xacK\xe5\xff<\x88+7\xef',
            {
                'ndef': ('*migr:test', 22),
                'props': {'bar': 'spam'},
                'tags': {}, 'tagprops': {}
            }
        )
    }
}

def getAssetBytes(*paths):
    fp = os.path.join(*paths)
    assert os.path.isfile(fp)
    with open(fp, 'rb') as f:
        byts = f.read()
    return byts

def getAssetJson(*paths):
    byts = getAssetBytes(*paths)
    obj = json.loads(byts.decode())
    return obj

# NOTE: This service is identical to the regression repo except the confdef has been
# updated for jsonschema and a new test value
class MigrSvcApi(s_stormsvc.StormSvc, s_cell.CellApi):
    _storm_svc_name = 'turtle'
    _storm_svc_pkgs = ({  # type: ignore
        'name': 'turtle',
        'version': (0, 0, 1),
        'commands': ({'name': 'newcmd', 'storm': '[ inet:fqdn=$lib.service.get($cmdconf.svciden).test() ]'},),
    },)

    async def test(self):
        return await self.cell.test()

class MigrStormsvc(s_cell.Cell):
    cellapi = MigrSvcApi
    confdefs = (
        ('myfqdn', {'type': 'string', 'default': 'snoop.io', 'description': 'A test fqdn'}),  # type: ignore
    )

    async def __anit__(self, dirn, conf=None):
        await s_cell.Cell.__anit__(self, dirn, conf=conf)
        self.myfqdn = self.conf.get('myfqdn')

    async def test(self):
        return self.myfqdn

class MigrationTest(s_t_utils.SynTest):

    @contextlib.asynccontextmanager
    async def _getTestMigrCore(self, conf):
        '''
        Use regression cortex as the base migration dirn.

        Args:
            conf (dict): Migration tool configuration

        Yields:
            (list): List of podes in the cortex
            (s_migr.Migrator): Migrator service object
            (tuple): test data, dest dirn, dest local layers, Migrator
        '''
        # get test data
        tdata = {}
        with self.getRegrDir('assets', REGR_VER) as assetdir:
            podesj = getAssetJson(assetdir, 'podes.json')
            ndj = getAssetJson(assetdir, 'nodedata.json')
            splicej = getAssetJson(assetdir, 'splices.json')
            splicepodesj = getAssetJson(assetdir, 'splicepodes.json')

        # strip out data we don't expect to migrate
        podesj = [p for p in podesj if p[0] not in NOMIGR_NDEF]
        ndj = [nd for nd in ndj if nd[0] not in NOMIGR_NDEF]

        tdata['podes'] = s_common.tuplify(podesj)
        tdata['nodedata'] = s_common.tuplify(ndj)

        tdata['splices'] = splicej
        tdata['splicepodes'] = splicepodesj

        # initialize migration tool
        with self.getRegrDir('cortexes', REGR_VER) as src:
            with self.getTestDir(copyfrom=conf.get('dest')) as dest:
                tconf = copy.deepcopy(conf)
                tconf['src'] = src
                tconf['dest'] = dest

                locallyrs = os.listdir(os.path.join(src, 'layers'))

                async with await s_migr.Migrator.anit(tconf) as migr:
                    yield tdata, dest, locallyrs, migr

    async def _checkStats(self, tdata, migr, locallyrs):
        '''
        Verify that the stats for what data has been migrated matches the test data
        '''
        tpodes = tdata['podes']
        tnodedata = tdata['nodedata']

        ipv4_cnt = len([x for x in tpodes if x[0][0] == 'inet:ipv4'])
        bytes_cnt = len([x for x in tpodes if x[0][0] == 'file:bytes'])
        tag_cnt = len([x for x in tpodes if x[0][0] == 'syn:tag'])

        nodedata_cnt = sum([len(x[1]) for x in tnodedata])

        ipv4_migr = [0, 0]
        bytes_migr = [0, 0]
        tag_migr = [0, 0]
        totnodes_migr = 0

        totnodedata_migr = 0

        for iden in locallyrs:
            stats = [log async for log in migr._migrlogGet('nodes', 'stat', f'{iden}:form')]
            self.gt(len(stats), 0)
            for stat in stats:
                skey = stat['key']
                sval = stat['val']  # (src_cnt, dest_cnt)

                if skey.endswith('inet:ipv4'):
                    ipv4_migr[0] += sval[0]
                    ipv4_migr[1] += sval[1]
                elif skey.endswith('file:bytes'):
                    bytes_migr[0] += sval[0]
                    bytes_migr[1] += sval[1]
                elif skey.endswith('syn:tag'):
                    tag_migr[0] += sval[0]
                    tag_migr[1] += sval[1]

            totnodes = [log async for log in migr._migrlogGet('nodes', 'stat', f'{iden}:totnodes')]
            totnodes_migr += totnodes[0]['val'][1]  # dest cnt

            totnodedata = [log async for log in migr._migrlogGet('nodedata', 'stat', f'{iden}:totnodes')]
            totnodedata_migr += totnodedata[0]['val'][1]  # dest cnt

        # check that sums from the layer migrations add up
        self.eq((ipv4_cnt, ipv4_cnt), ipv4_migr)
        self.eq((bytes_cnt, bytes_cnt), bytes_migr)
        self.eq((tag_cnt, tag_cnt), tag_migr)
        self.eq(len(tpodes), totnodes_migr)

        self.eq(nodedata_cnt, totnodedata_migr)

    async def _checkCore(self, core, tdata, nodesonly=False):
        '''
        Verify data in the migrated to 2.x.x cortex.
        '''

        # test validation data
        tpodes = tdata['podes']
        tnodedata = tdata['nodedata']

        # check all nodes (removing empty nodedata key)
        nodes = await core.nodes('.created -meta:source:name=test')

        podes = [n.pack(dorepr=True) for n in nodes]
        podes = [(p[0], {k: v for k, v in p[1].items() if k != 'nodedata'}) for p in podes]
        self.gt(len(podes), 0)

        try:
            self.eq(podes, tpodes)
        except AssertionError:
            # print a more useful diff on error
            notincore = list(itertools.filterfalse(lambda x: x in podes, tpodes))
            notintest = list(itertools.filterfalse(lambda x: x in tpodes, podes))
            self.eq([], notincore)
            self.eq([], notintest)
            raise

        if not nodesonly:
            nodedata = []
            for n in nodes:
                nodedata.append([n.ndef, [nd async for nd in n.iterData()]])
            nodedata = s_common.tuplify(nodedata)
            try:
                self.eq(nodedata, tnodedata)
            except AssertionError:
                # print a more useful diff on error
                notincore = list(itertools.filterfalse(lambda x: x in nodedata, tnodedata))
                notintest = list(itertools.filterfalse(lambda x: x in tnodedata, nodedata))
                self.eq([], notincore)
                self.eq([], notintest)
                raise

        # manually check node subset
        self.len(1, await core.nodes('inet:ipv4=1.2.3.4'))
        self.len(2, await core.nodes('inet:dns:a:ipv4=1.2.3.4'))

        if not nodesonly:
            # check that triggers are active
            await core.auth.getUserByName('root')
            self.len(5, await core.nodes('syn:trigger'))
            self.len(5, await core.callStorm('return ($lib.trigger.list())'))
            tnodes = await core.nodes('[ inet:ipv4=9.9.9.9 ]')
            self.nn(tnodes[0].tags.get('trgtag'))

            # make sure trigger conditions don't exist with a None value
            conds = ['form', 'prop', 'tag']
            trigs = await core.view.listTriggers()
            self.true(all([trig[1].pack().get(x, 'empty') is not None for trig in trigs for x in conds]))

    async def _checkSplices(self, core0, tdata):
        '''
        Validate splices by feeding into a fresh cortex and evaluating as if it was the original migrated cortex
        '''
        with self.getTestDir() as dirn:
            conf = {
                'layer:lmdb:map_async': True,
                'provenance:en': False,
                'nexslog:en': False,
                'layers:logedits': False,
            }
            async with self.getTestCore(dirn=dirn) as core1:
                # give the new core the same hive for datamodel, triggers, etc.
                tree0 = await core0.hive.saveHiveTree()
                await core1.loadHiveTree(copy.deepcopy(tree0))

            async with self.getTestCore(dirn=dirn, conf=conf) as core1:
                # primary layer
                lyr00 = core0.getLayer()
                lyr01 = core1.getLayer()

                nes0 = [nodeedits for offs, nodeedits in lyr00.nodeeditlog.iter(0)]
                [await lyr01.storNodeEdits(nes[0], None) for nes in nes0]
                self.lt(len(nes0), tdata['splices'][lyr00.iden]['nextoffs'])

                # secondary layer
                lyr10 = [lyr for lyr in core0.listLayers() if lyr.iden != lyr00.iden][0]
                lyr11def = await core1.addLayer()
                lyr11 = core1.getLayer(lyr11def['iden'])

                nes1 = [nodeedits for offs, nodeedits in lyr10.nodeeditlog.iter(0)]
                [await lyr11.storNodeEdits(nes[0], None) for nes in nes1]
                self.lt(len(nes1), tdata['splices'][lyr10.iden]['nextoffs'])

                await self._checkCore(core1, tdata, nodesonly=True)

    async def _checkCell(self, core):
        '''
        Check cell data *except* auth
        '''

        rootuser = core.auth.rootuser.iden

        self.eq(s_version.version, core.cellinfo.get('cortex:version'))

        for view in core.views.values():
            self.eq(rootuser, view.info.get('creator'))

        for lyr in core.layers.values():
            self.true(lyr.lockmemory)
            self.false(lyr.readonly)
            self.true(lyr.logedits)
            self.eq(rootuser, lyr.layrinfo.get('creator'))
            self.eq(s_modelrev.maxvers, await lyr.getModelVers())

    async def _checkAuth(self, core):
        defview = await core.hive.get(('cellinfo', 'defaultview'))
        deflyr = core.getLayer().iden
        secview = [k for k in core.views.keys() if k != defview][0]
        seclyr = core.views[secview].layers[0].iden
        self.ne(deflyr, seclyr)  # check to make sure we got the second layer

        # data check auth layout (users, passwords, rules, admin, etc.)
        self.sorteq(list(core.auth.usersbyname.keys()), ['root', 'fred', 'bobo'])
        self.sorteq(list(core.auth.rolesbyname.keys()), ['all', 'cowboys', 'ninjas', 'friends'])

        self.len(15, core.auth.authgates)  # 2 views, 2 layers, 5 triggers, 1 cortex, 3 queues, 2 crons
        self.len(2, [iden for iden, gate in core.auth.authgates.items() if gate.type == 'view'])
        self.len(2, [iden for iden, gate in core.auth.authgates.items() if gate.type == 'layer'])
        self.len(1, [iden for iden, gate in core.auth.authgates.items() if gate.type == 'cortex'])
        self.len(3, [iden for iden, gate in core.auth.authgates.items() if gate.type == 'queue'])
        self.len(2, [iden for iden, gate in core.auth.authgates.items() if gate.type == 'cronjob'])

        # user attributes
        root = core.auth.usersbyname['root']
        self.true(root.info.get('admin'))
        self.sorteq(['all'], [core.auth.rolesbyiden[riden].name for riden in root.info.get('roles')])
        self.len(0, root.info.get('rules'))

        bobo = core.auth.usersbyname['bobo']
        self.false(bobo.info.get('admin'))
        self.sorteq(['all', 'friends'], [core.auth.rolesbyiden[riden].name for riden in bobo.info.get('roles')])
        boborules = [(True, ('node', 'tag', 'add', 'bobotag')), (True, ('queue', 'get')), (True, ('queue', 'boboq')),
                     (True, ('queue', 'put'))]
        self.sorteq(boborules, bobo.info.get('rules', []))

        fred = core.auth.usersbyname['fred']
        self.false(fred.info.get('admin'))
        self.sorteq(['all', 'ninjas'], [core.auth.rolesbyiden[riden].name for riden in fred.info.get('roles')])
        fredrules = [(True, ('node', 'tag', 'add', 'trgtag')), (True, ('node', 'tag', 'add', 'trgtagdel')),
                     (True, ('trigger', 'add')), (True, ('trigger', 'get')), (True, ('queue', 'get')),
                     (True, ('queue', 'add')), (True, ('cron',))]
        self.sorteq(fredrules, fred.info.get('rules', []))

        # role attributes
        friends = core.auth.rolesbyname['friends']
        friendrules = [(True, ('queue', 'fredq', 'get')), (True, ('cron', 'get'))]
        self.sorteq(friendrules, friends.info.get('rules', []))
        friendrules_seclyr = [(True, ('node', 'add')), (True, ('node', 'prop', 'set')), (True, ('layer', 'lift'))]
        self.sorteq(friendrules_seclyr, friends.authgates[seclyr].get('rules'))

        # load vals for user perm tests
        tagtrg = (await core.nodes('syn:trigger:cond=tag:add'))[0].ndef[1]
        nodetrg = (await core.nodes('syn:trigger:cond=node:add'))[0].ndef[1]

        crons = await core.listCronJobs()
        fredcron = [c for c in crons if c['creator'] == fred.iden][0]
        bobocron = [c for c in crons if c['creator'] == bobo.iden][0]

        # user permissions
        # bobo
        # - read main view
        # - read/write to forked view via role
        # - no access to triggers
        # - can add/del bobotag but not trgtag
        # - has queue get/put but not queue add
        # - cron get via friends role
        async with core.getLocalProxy(user='bobo') as proxy:
            self.gt(await proxy.count('inet:ipv4'), 0)
            await self.asyncraises(s_exc.AuthDeny, proxy.count('[inet:ipv4=10.10.10.10]'))

            self.eq(5, await proxy.count('syn:trigger'))
            await self.asyncraises(s_exc.StormRuntimeError, proxy.count(f'trigger.del {tagtrg}'))
            await self.asyncraises(s_exc.StormRuntimeError, proxy.count(f'trigger.del {nodetrg}'))
            trigadd = 'trigger.add node:add --form file:bytes --query {[+#bobotag]}'
            await self.asyncraises(s_exc.AuthDeny, proxy.count(f'{trigadd}'))
            self.eq(5, await proxy.count('syn:trigger'))

            self.gt(await proxy.count('inet:ipv4 [+#bobotag]'), 0)
            await self.asyncraises(s_exc.AuthDeny, proxy.count('#trgtag [-#trgtag]'))
            await self.asyncraises(s_exc.AuthDeny, proxy.count('inet:ipv4 [+#newbobotag]'))

            self.gt(await proxy.count('inet:ipv4', opts={'view': secview}), 0)
            self.gt(await proxy.count('[inet:ipv4=10.9.10.1]', opts={'view': secview}), 0)

            await self.asyncraises(s_exc.AuthDeny, proxy.count(f'queue.add newboboq'))

            await proxy.count(f'$q = $lib.queue.get(boboq) inet:ipv4=1.2.3.4 $q.put($node.repr())')
            self.eq(1, await proxy.count(f'$q = $lib.queue.get(boboq) ($offs, $ipv4) = $q.get(0) inet:ipv4=$ipv4'))

            await proxy.count(f'$q = $lib.queue.get(fredq) inet:ipv4=1.2.3.4 $q.put($node.repr())')
            self.eq(1, await proxy.count(f'$q = $lib.queue.get(fredq) ($offs, $ipv4) = $q.get(0) inet:ipv4=$ipv4'))

            self.len(2, await proxy.listCronJobs())
            await proxy.count(f'cron.disable {bobocron["iden"]}')

            await self.asyncraises(s_exc.AuthDeny, proxy.count('cron.add --hour +8 {file:bytes}'))
            await self.asyncraises(s_exc.StormRuntimeError, proxy.count(f'cron.del {fredcron["iden"]}'))
            await self.asyncraises(s_exc.StormRuntimeError, proxy.count(f'cron.disable {fredcron["iden"]}'))

        # fred
        # - read to main view
        # - read access to forked view via rule
        # - get/add all triggers, but can only delete his own
        # - can add/del trgtag but not bobotag
        # - has queue add/get, and admin on fredq
        # - full cron rights
        async with core.getLocalProxy(user='fred') as proxy:
            self.gt(await proxy.count('inet:ipv4'), 0)
            await self.asyncraises(s_exc.AuthDeny, proxy.count('[inet:ipv4=10.10.10.10]'))

            self.eq(5, await proxy.count('syn:trigger'))

            await self.asyncraises(s_exc.AuthDeny, proxy.count(f'trigger.del {tagtrg}'))
            await proxy.count(f'trigger.del {nodetrg}')
            self.eq(4, await proxy.count('syn:trigger'))

            await proxy.count('trigger.add node:add --form file:bytes --query {[+#trgtag]}')
            self.eq(5, await proxy.count('syn:trigger'))

            self.gt(await proxy.count('inet:ipv4 [+#trgtag]'), 0)
            await self.asyncraises(s_exc.AuthDeny, proxy.count('#trgtag [-#bobotag]'))
            await self.asyncraises(s_exc.AuthDeny, proxy.count('inet:ipv4 [+#newfredtag]'))

            self.gt(await proxy.count('inet:ipv4', opts={'view': secview}), 0)
            await self.asyncraises(s_exc.AuthDeny, proxy.count('[inet:ipv4=10.9.10.1]', opts={'view': secview}))

            await proxy.count(f'queue.add newfredq')

            await proxy.count(f'$q = $lib.queue.get(fredq) inet:ipv4=9.9.9.9 $q.put($node.repr())')
            self.eq(1, await proxy.count(f'$q = $lib.queue.get(fredq) ($offs, $ipv4) = $q.get(0) inet:ipv4=$ipv4'))

            await self.asyncraises(s_exc.AuthDeny,
                                   proxy.count(f'$q = $lib.queue.get(boboq) inet:ipv4=1.2.3.4 $q.put($node.repr())'))
            await self.asyncraises(s_exc.AuthDeny,
                                   proxy.count(f'$q = $lib.queue.get(rootq) inet:ipv4=1.2.3.4 $q.put($node.repr())'))

            self.len(2, await proxy.listCronJobs())
            await proxy.count(f'cron.del {bobocron["iden"]}')
            self.len(1, await proxy.listCronJobs())

            await proxy.count(f'cron.disable {fredcron["iden"]}')

            await proxy.count('cron.add --hour +8 {file:bytes}')
            self.len(2, await proxy.listCronJobs())

        # root
        # - read/write to main view
        # - read/write to forked view
        # - get/del triggers
        # - full rights to all queues
        # - full rights to cronjobs
        async with core.getLocalProxy(user='root') as proxy:
            self.gt(await proxy.count('inet:ipv4'), 0)
            self.gt(await proxy.count('[inet:ipv4=10.10.10.11]'), 0)

            self.eq(5, await proxy.count('syn:trigger'))
            await proxy.count(f'$lib.trigger.del({tagtrg})')
            self.eq(4, await proxy.count('syn:trigger'))

            self.gt(await proxy.count('#faz.baz'), 0)
            self.eq(0, await proxy.count('#trgtagdel'))
            await proxy.count('#faz.baz -#faz.baz.* | [-#faz.baz]')
            self.gt(await proxy.count('#trgtagdel'), 0)

            await proxy.count('inet:ipv4=7.7.7.7 | delnode')
            self.eq(0, await proxy.count('inet:ipv4=7.7.7.7'))
            await proxy.count(f'file:bytes | limit 1 | delnode')
            self.eq(1, await proxy.count('inet:ipv4=7.7.7.7'))

            self.eq(0, await proxy.count('#proptrgtag'))
            await proxy.count(f'file:bytes | limit 1 | [:name=foofaz]')
            self.eq(1, await proxy.count('file:bytes:name=foofaz +#proptrgtag'))

            self.gt(await proxy.count('inet:ipv4', opts={'view': secview}), 0)
            self.gt(await proxy.count('[inet:ipv4=10.9.10.1]', opts={'view': secview}), 0)

            await proxy.count(f'$q = $lib.queue.get(boboq) inet:ipv4=9.9.9.9 $q.put($node.repr())')
            self.eq(1, await proxy.count(f'$q = $lib.queue.get(boboq) ($offs, $ipv4) = $q.get(0) inet:ipv4=$ipv4'))

            await proxy.count(f'$q = $lib.queue.get(fredq) inet:ipv4=9.9.9.9 $q.put($node.repr())')
            self.eq(1, await proxy.count(f'$q = $lib.queue.get(boboq) ($offs, $ipv4) = $q.get(0) inet:ipv4=$ipv4'))

            await proxy.count(f'$q = $lib.queue.get(rootq) inet:ipv4=9.9.9.9 $q.put($node.repr())')
            self.eq(1, await proxy.count(f'$q = $lib.queue.get(rootq) ($offs, $ipv4) = $q.get(0) inet:ipv4=$ipv4'))

            crons = await proxy.listCronJobs()
            self.len(2, crons)

            await proxy.count('cron.add --hour +8 {inet:email}')
            self.len(3, await proxy.listCronJobs())

            await proxy.count(f'cron.del {fredcron["iden"]}')
            self.len(2, await proxy.listCronJobs())

    async def test_migr_nexus(self):
        conf = {
            'src': None,
            'dest': None,
            'addmode': 'nexus',
            'migrops': None,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr.migrate()

            await self._checkStats(tdata, migr, locallyrs)

            # test dump errors
            dumpf = await migr.dumpErrors()
            self.nn(dumpf)

            errs = []
            with open(dumpf, 'rb') as fd:
                errs = s_msgpack.un(fd.read())

            self.len(2, errs)
            for err in errs:
                err['val']['node'][1]['props'].pop('.created')
            self.isin(MIGR_ERR, errs)

            await migr.fini()

            # startup 2.x core
            async with await s_cortex.Cortex.anit(dest, conf={'nexslog:en': True}) as core:
                # check that nexus root has offsets from migration
                self.gt(core.nexsroot.nexslog.index(), 1)

                # check core data
                await self._checkSplices(core, tdata)
                await self._checkCore(core, tdata)
                await self._checkCell(core)
                await self._checkAuth(core)

    async def test_migr_nexusoff(self):
        conf = {
            'src': None,
            'dest': None,
            'safetyoff': True,
            'migrops': None,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr.migrate()

            await self._checkStats(tdata, migr, locallyrs)

            await migr.fini()

            # startup 2.x core
            async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                # check that nexus root has *no* offsets from migration
                self.eq(core.nexsroot.nexslog.index(), 0)

                # check core data
                await self._checkSplices(core, tdata)
                await self._checkCore(core, tdata)
                await self._checkCell(core)
                await self._checkAuth(core)

    async def test_migr_editor(self):
        conf = {
            'src': None,
            'dest': None,
            'addmode': 'editor',
            'migrops': None,
            'fairiter': 1,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr.migrate()

            await self._checkStats(tdata, migr, locallyrs)

            await migr.fini()

            # startup 2.x core
            async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                # check that nexus root has *no* offsets from migration
                self.eq(core.nexsroot.nexslog.index(), 0)

                # check core data
                await self._checkSplices(core, tdata)
                await self._checkCore(core, tdata)
                await self._checkCell(core)
                await self._checkAuth(core)

    async def test_migr_chkpnt(self):
        conf = {
            'src': None,
            'dest': None,
            'addmode': 'nexus',
            'fromlast': True,
            'nodelim': 25,
            'migrops': None,
            'fairiter': 1,
        }

        async with self._getTestMigrCore(conf) as (tdata0, dest0, locallyrs0, migr0):
            migr0.savechkpnt = 7
            await migr0.migrate()

            logs = [log async for log in migr0._migrlogGet('nodes', 'chkpnt')]
            self.len(1, logs)
            self.eq(25, logs[0]['val'][1])  # last chkpnt was at nodelim
            lyriden = logs[0]['key']

            logs = [log async for log in migr0._migrlogGet('nodes', 'stat', f'{lyriden}:totnodes')]
            self.len(1, logs)
            self.eq(24, logs[0]['val'][0])
            self.ge(24, logs[0]['val'][1])

            await migr0.fini()

            # finish the migration
            conf['dest'] = dest0
            conf['nodelim'] = None

            async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
                await migr.migrate()

                log = await migr._migrlogGetOne('nodes', 'chkpnt', lyriden)
                self.eq(len(tdata['podes']), log['val'][1] - 2)  # -1 since next node, -1 for err not in tdata

                await self._checkStats(tdata, migr, locallyrs)

                await migr.fini()

                # startup 2.x core
                async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                    # check core data
                    await self._checkSplices(core, tdata)
                    await self._checkCore(core, tdata)
                    await self._checkCell(core)
                    await self._checkAuth(core)

    async def test_migr_restart(self):
        conf = {
            'src': None,
            'dest': None,
            'migrops': None,
            'fairiter': 1,
        }

        async with self._getTestMigrCore(conf) as (tdata0, dest0, locallyrs0, migr0):
            # add a boot.yaml file
            s_common.yamlsave({'auth:admin': 'root:root'}, migr0.src, 'boot.yaml')

            await migr0.migrate()

            await self._checkStats(tdata0, migr0, locallyrs0)

            # get layer offsets
            offslogs0 = []
            for iden in locallyrs0:
                offslogs0 += [log async for log in migr0._migrlogGet('nodes', 'nextoffs', iden)]

            self.len(2, offslogs0)  # one entry for each layer
            self.gt(offslogs0[0]['val'][0], 0)
            self.gt(offslogs0[1]['val'][0], 0)

            # check the saved offset file
            offsyaml = s_common.yamlload(dest0, 'migration', 'lyroffs.yaml')
            offslogdict = {x['key']: {'nextoffs': x['val'][0], 'created': x['val'][1]} for x in offslogs0}
            self.eq(offsyaml, offslogdict)

            # check the saved version file
            versyaml = s_common.yamlload(dest0, 'migration', 'migrvers.yaml')
            self.eq([(0, 1, 3), (0, 1, 3)], [tuple(v) for v in versyaml.get('src:model', {}).values()])
            self.eq((0, 1, 56), versyaml.get('src:cortex'))
            self.eq({s_version.version}, {tuple(versyaml.get('dest:cortex').get(op)) for op in migr0.migrops})

            await migr0.fini()

            # run migration again
            conf['dest'] = dest0

            async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
                # check form counts
                fcntprnt = await migr.formCounts()
                self.len(2, fcntprnt)
                fnum = len([x for x in tdata['podes'] if x[0][0] == 'inet:fqdn'])
                self.isin(f'inet:fqdn{fnum}{fnum}0', '_'.join(fcntprnt).replace(' ', ''))

                # check that destination is populated before starting migration
                iden = locallyrs[0]
                lyrslab = os.path.join(dest, 'layers', iden, 'layer_v2.lmdb')
                self.true(os.path.exists(lyrslab))

                # add a boot.yaml file
                s_common.yamlsave({'auth:admin': 'root:root'}, migr.src, 'boot.yaml')

                await migr.migrate()

                await self._checkStats(tdata, migr, locallyrs)

                # get layer offsets
                offslogs = []
                for iden in locallyrs:
                    offslogs += [log async for log in migr._migrlogGet('nodes', 'nextoffs', iden)]

                self.len(2, offslogs)  # one entry for each layer
                self.gt(offslogs[0]['val'][0], 0)
                self.gt(offslogs[1]['val'][0], 0)
                self.gt(offslogs[0]['val'][1], offslogs0[0]['val'][1])  # timestamp should be updated
                self.gt(offslogs[1]['val'][1], offslogs0[1]['val'][1])

                # check the saved offset file
                offsyaml = s_common.yamlload(dest, 'migration', 'lyroffs.yaml')
                self.eq(offsyaml, {x['key']: {'nextoffs': x['val'][0], 'created': x['val'][1]} for x in offslogs})

                # check the saved version file
                versyaml = s_common.yamlload(dest, 'migration', 'migrvers.yaml')
                self.eq([(0, 1, 3), (0, 1, 3)], [tuple(v) for v in versyaml.get('src:model', {}).values()])
                self.eq((0, 1, 56), versyaml.get('src:cortex'))
                self.eq({s_version.version}, {tuple(versyaml.get('dest:cortex').get(op)) for op in migr.migrops})

                # check the boot.yaml
                self.false(os.path.exists(os.path.join(dest, 'boot.yaml')))
                self.true(os.path.exists(os.path.join(dest, 'migration', 'boot.yaml')))
                self.len(1, [log async for log in migr._migrlogGet('cell', 'error')])

                await migr.fini()

                # startup 2.x core
                async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                    # check core data
                    await self._checkSplices(core, tdata)
                    await self._checkCore(core, tdata)
                    await self._checkCell(core)
                    await self._checkAuth(core)

    async def test_migr_assvr_defaults(self):
        '''
        Test that migration service is being properly initialized from default cmdline args.
        '''
        with self.getRegrDir('cortexes', REGR_VER) as src:
            # sneak in test for missing splice slab - no impact to migration
            for root, dirs, _ in os.walk(src, topdown=True):
                for dname in dirs:
                    if dname == 'splices.lmdb':
                        shutil.rmtree(os.path.join(root, dname))

            cellconf = {}
            s_common.yamlsave(cellconf, src, 'cell.yaml')

            # check defaults
            with self.getTestDir() as dest, self.withSetLoggingMock():
                argv = [
                    '--src', src,
                    '--dest', dest,
                ]

                async with await s_migr.main(argv) as migr:
                    self.true(migr.isfini)
                    self.eq(migr.src, src)
                    self.eq(migr.dest, dest)
                    self.sorteq(migr.migrops, s_migr.ALL_MIGROPS)
                    self.eq(migr.addmode, 'nonexus')
                    self.eq(migr.editbatchsize, 100)
                    self.eq(migr.fairiter, 100)
                    self.none(migr.nodelim)
                    self.false(migr.safetyoff)
                    self.false(migr.fromlast)
                    self.false(migr.srcdedicated)
                    self.false(migr.destdedicated)

                    # check the saved offset file
                    offsyaml = s_common.yamlload(dest, 'migration', 'lyroffs.yaml')
                    self.true(all(v['nextoffs'] == 0 for v in offsyaml.values()))

                # startup 2.x core
                async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                    nodes = await core.nodes('inet:ipv4=1.2.3.4')
                    self.len(1, nodes)
                    nodes = await core.nodes('[inet:ipv4=9.9.9.9]')
                    self.len(1, nodes)

                    for lyr in core.layers.values():
                        self.false(lyr.lockmemory)

    async def test_migr_assvr_opts(self):
        '''
        Test that migration service is being properly initialized from user cmdline args.
        '''
        with self.getRegrDir('cortexes', REGR_VER) as src:
            # check user opts
            with self.getTestDir() as destp, self.withSetLoggingMock():
                dest = os.path.join(destp, 'woot')  # verify svc is creating dir if it doesn't exist

                argv = [
                    '--src', src,
                    '--dest', dest,
                    '--add-mode', 'editor',
                    '--nodelim', '1000',
                    '--edit-batchsize', '1000',
                    '--fair-iter', '5',
                    '--src-dedicated',
                    '--dest-dedicated',
                    '--from-last',
                    '--safety-off',
                    '--migr-ops', 'dirn', 'dmodel', 'cell',
                ]

                async with await s_migr.main(argv) as migr:
                    self.eq(migr.src, src)
                    self.eq(migr.dest, dest)
                    self.sorteq(migr.migrops, ['dirn', 'dmodel', 'cell'])
                    self.eq(migr.addmode, 'editor')
                    self.eq(migr.editbatchsize, 1000)
                    self.eq(migr.fairiter, 5)
                    self.eq(migr.nodelim, 1000)
                    self.true(migr.safetyoff)
                    self.true(migr.fromlast)
                    self.true(migr.srcdedicated)
                    self.true(migr.destdedicated)

                argv = [
                    '--src', src,
                    '--dest', dest,
                    '--form-counts',
                ]

                outp = self.getTestOutp()
                async with await s_migr.main(argv, outp=outp) as migr:
                    self.true(outp.expect('Form counts for layer', throw=False))

                argv = [
                    '--src', src,
                    '--dest', dest,
                    '--dump-errors',
                ]

                outp = self.getTestOutp()
                async with await s_migr.main(argv, outp=outp) as migr:
                    self.true(outp.expect('Dump file located at', throw=False))

    async def test_migr_errconf(self):
        with self.getRegrDir('cortexes', REGR_VER) as src:
            with self.getTestDir() as dest:

                conf = {
                    'src': src,
                    'dest': None,
                }

                async with await s_migr.Migrator.anit(conf) as migr:
                    res = await migr.formCounts()
                    fullprnt = ' '.join(res)
                    locallyrs = os.listdir(os.path.join(src, 'layers'))
                    self.isin(locallyrs[0], fullprnt)
                    self.isin(locallyrs[1], fullprnt)

                    await self.asyncraises(Exception, migr.migrate())

                conf = {
                    'src': src,
                    'dest': dest,
                    'addmode': 'foobar',
                    'editbatchsize': 3,
                }
                await self.asyncraises(Exception, s_migr.Migrator.anit(conf))

                conf = {
                    'src': src,
                    'dest': dest,
                    'addmode': 'nexus',
                    'editbatchsize': 3,
                }

                async with await s_migr.Migrator.anit(conf) as migr:
                    migr.addmode = 'foobar'
                    await migr.migrate()

                    errs = [log async for log in migr._migrlogGet('nodes', 'error')]
                    self.isin('Unrecognized addmode foobar', [x.get('val', {}).get('mesg') for x in errs])

                    # migration dir gets deleted
                    shutil.rmtree(os.path.join(dest, 'migration'))
                    res = await migr.dumpErrors()
                    self.none(res)

                # startup 2.x core - with no nodes
                async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                    nodes = await core.nodes('inet:ipv4=1.2.3.4')
                    self.len(0, nodes)

    async def test_migr_missingidens(self):
        with self.getRegrDir('cortexes', REGR_VER) as src:
            with self.getTestDir() as dest:

                conf = {
                    'src': src,
                    'dest': dest,
                }

                async with await s_migr.Migrator.anit(conf) as migr:
                    await migr._migrDirn()
                    await migr._initStors()

                    # replace user idens in hive
                    root = None
                    fred = None
                    newroot = s_common.guid()
                    newfred = s_common.guid()

                    users = await migr.hive.open(('auth', 'users'))
                    usersd = await users.dict()
                    for uiden, uname in usersd.items():
                        if uname == 'root':
                            root = uiden
                        elif uname == 'fred':
                            fred = uiden

                    await usersd.set(fred, None)  # no username

                    await migr.hive.rename(('auth', 'users', root), ('auth', 'users', newroot))
                    await migr.hive.rename(('auth', 'users', fred), ('auth', 'users', newfred))

                    # by not migrating the actual crons and triggers, the auth migration
                    # will create new auth gates but won't have a matching user
                    migr.migrops = [op for op in s_migr.ALL_MIGROPS if op not in ('dirn', 'cron')]

                    await migr.migrate()

                # startup 2.x core
                async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                    nodes = await core.nodes('inet:ipv4=1.2.3.4')
                    self.len(1, nodes)
                    nodes = await core.nodes('[inet:ipv4=9.9.9.9]')
                    self.len(1, nodes)

    async def test_migr_celliden(self):
        conf = {
            'src': None,
            'dest': None,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr._migrDirn()
            await migr._initStors()

            guidpath = os.path.join(dest, 'cell.guid')

            with open(guidpath, 'r') as fd:
                iden0 = fd.read().strip()

            self.true(await migr._chkValid())
            self.eq(iden0, (await migr._migrlogGetOne('chkvalid', 'iden', 'src:cortex'))['val'])

            # remove cell.guid
            os.remove(guidpath)

            self.false(await migr._chkValid())
            self.none((await migr._migrlogGetOne('chkvalid', 'iden', 'src:cortex'))['val'])

            # replace
            with open(guidpath, 'w') as fd:
                fd.write(iden0)

            self.true(await migr._chkValid())

            # set guid to a layer iden
            iden1 = locallyrs[0]
            with open(guidpath, 'w') as fd:
                fd.write(iden1)

            self.false(await migr._chkValid())
            self.eq(iden1, (await migr._migrlogGetOne('chkvalid', 'iden', 'src:cortex'))['val'])

    async def test_migr_yamlmod(self):
        with self.getRegrDir('cortexes', REGR_VER) as src:
            with self.getTestDir() as dest:
                locallyrs = os.listdir(os.path.join(src, 'layers'))

                mods = {'modules': ['synapse.tests.utils.TestModule']}
                s_common.yamlsave(mods, src, 'cell.yaml')
                s_common.yamlsave({}, src, 'boot.yaml')  # empty, but will still move

                conf = {
                    'src': src,
                    'dest': dest,
                    'migrops': ['dmodel', 'hivelyr', 'cell', 'nodes'],
                    'editbatchsize': 1,
                    'nodelim': 2,
                }

                async with await s_migr.Migrator.anit(conf) as migr:
                    # test that we can skip dirn migration if it already exists
                    migr.migrops.append('dirn')
                    await migr._migrDirn()
                    del migr.migrops[-1]

                    # verify that test:int was loaded and migrated
                    await migr.migrate()
                    stats = []
                    for iden in locallyrs:
                        statkey = f'{iden}:form:test:int'
                        stats.extend([log async for log in migr._migrlogGet('nodes', 'stat', statkey)])

                    self.eq((1, 1), stats[0]['val'])

                    self.false(os.path.exists(os.path.join(dest, 'boot.yaml')))
                    self.true(os.path.exists(os.path.join(dest, 'migration', 'boot.yaml')))

    async def test_migr_stormsvc(self):
        conf = {
            'src': None,
            'dest': None,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr.migrate()
            await migr.fini()

            # startup 2.x core
            async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                # check core data
                await self._checkCore(core, tdata)

                # add in service
                svcpath = os.path.join(dest, 'stormsvc')
                async with await MigrStormsvc.anit(svcpath) as svc:
                    svc.dmon.share('turtle', svc)
                    root = await svc.auth.getUserByName('root')
                    await root.setPasswd('root')
                    info = await svc.dmon.listen('tcp://127.0.0.1:0/')
                    host, port = info

                    await core.nodes(f'service.add turtle tcp://root:root@127.0.0.1:{port}/turtle')
                    await core.nodes('$lib.service.wait(turtle)')

                    await core.nodes('newcmd')
                    self.len(1, await core.nodes('inet:fqdn=snoop.io'))

    async def test_migr_invalid(self):
        conf = {
            'src': None,
            'dest': None,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            os.remove(os.path.join(migr.src, 'cell.yaml'))
            await migr._migrDirn()
            await migr._initStors()

            # modify a layer to be of type=remote in the hive
            lyrnode = await migr.hive.open(('cortex', 'layers', locallyrs[0]))
            layrinfo = await lyrnode.dict()
            await layrinfo.set('type', 'remote')

            migr.migrops = [op for op in migr.migrops if op != 'dirn']

            await migr.migrate()
            await migr.fini()

            # An error will be generated and migration halted
            # so we can check that the cortex is not startable as 2.x.x
            await self.asyncraises(s_exc.BadStorageVersion, s_cortex.Cortex.anit(dest, conf=None))

    async def test_migr_badmodelvers(self):
        with self.getRegrDir('cortexes', REGR_VER_OLD) as src:
            with self.getTestDir() as dest:
                conf = {
                    'src': src,
                    'dest': dest,
                }

                async with await s_migr.Migrator.anit(conf) as migr:
                    await self.asyncraises(Exception, migr.migrate())

    async def test_migr_splices(self):
        conf = {
            'src': None,
            'dest': None,
            'migrops': [op for op in s_migr.ALL_MIGROPS if op not in ('nodes', 'splices')],
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr.migrate()

            path = os.path.join(migr.dest, migr.migrdir, 'splices')
            conf = {
                'queue_size': 15,  # make artificially small to simulate read waiting
            }
            migr.migrsplices = await s_migr.MigrSplices.anit(path, conf=conf)
            migr.onfini(migr.migrsplices)

            migr.migrsplices.model.update(migr.model.getModelDict())
            await migr._migrSplices(locallyrs[0])
            await migr._migrSplices(locallyrs[1])

            await migr.fini()

            # startup 2.x core
            async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                # we should be able to iterate over the nodeedits, store them,
                # and bring the cortex to the same state as a node migration
                # this will generate a duplicate set of nodeedits in the layer
                self.len(0, await core.nodes('.created -meta:source:name=test'))
                self.eq(0, await core.count('.created'))

                lyr0 = core.layers[locallyrs[0]]
                nes0 = [nodeedits for offs, nodeedits in lyr0.nodeeditlog.iter(0)]
                [await lyr0.storNodeEdits(nes[0], None) for nes in nes0]

                lyr1 = core.layers[locallyrs[1]]
                nes1 = [nodeedits for offs, nodeedits in lyr1.nodeeditlog.iter(0)]
                [await lyr1.storNodeEdits(nes[0], None) for nes in nes1]

                await self._checkCore(core, tdata, nodesonly=True)

    async def test_migr_splice_errs(self):
        conf = {
            'src': None,
            'dest': None,
            'migrops': [op for op in s_migr.ALL_MIGROPS if op != 'splices'],
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr.migrate()

            path = os.path.join(migr.dest, migr.migrdir, 'splices')
            conf = {
                'queue_size': 15,
                'err_lim': 100,
            }
            migr.migrsplices = await s_migr.MigrSplices.anit(path, conf=conf)
            migr.onfini(migr.migrsplices)

            # load a partial datamodel which will generate errors
            # also add a few forms / props to old model chk
            migr.migrsplices.model = {
                'forms': {'inet:fqdn': {'stortype': 1, 'props': {}}},
                'tagprops': {},
            }
            migr.migrsplices.form_chk['inet:ipv4'] = 0
            migr.migrsplices.prop_chk['inet:fqdn:host'] = 0

            await migr._migrSplices(locallyrs[0])
            await migr._migrSplices(locallyrs[1])

            # remove the genr and run src iteration
            migr.migrsplices.src_genrs[locallyrs[1]] = None
            migr.migrsplices._queues[locallyrs[1]].isfini = False
            task = migr.migrsplices.schedCoro(migr.migrsplices._srcPullLyrSplices(locallyrs[1]))
            await asyncio.wait_for(task, timeout=10)
            res = task.result()
            self.false(res.get('status'))

            await migr.fini()

            # startup 2.x core and check that partial nodeedits were stored
            async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                lyr0 = core.getLayer()  # primary layer

                # the only nodeedits should be for file:byte node adds
                exp = len([x for x in tdata['podes'] if x[0][0] == 'inet:fqdn'])
                offs = await lyr0.getNodeEditOffset()
                self.eq(exp, offs)
                nes = [nodeedits for offs, nodeedits in lyr0.nodeeditlog.iter(0)]
                self.len(exp, nes)

                # even though there were splice migration errors, nodes and nodedata are still intact
                await self._checkCore(core, tdata)

    async def test_migr_cell(self):
        conf = {
            'src': None,
            'dest': None,
            'migrops': [op for op in s_migr.ALL_MIGROPS if op != 'cell'],
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr.migrate()
            await migr.fini()

            await self.asyncraises(s_exc.BadConfValu, s_cortex.Cortex.anit(dest, conf=None))

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            migr.migrops.append('cell')
            os.remove(os.path.join(migr.src, 'cell.yaml'))

            await migr.migrate()
            await migr.fini()

            # startup 2.x core
            async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                # check core data
                await self._checkCore(core, tdata)

    async def test_migr_auth(self):
        '''
        Smaller unit tests for auth migration
        '''
        migrauth = s_migr.MigrAuth(None, None, None, None, None)

        rules_in = [
            (True, ('foo', 'baz')),
            (True, ('storm',)),
            (True, ('storm', 'queue', 'get')),
            (True, ('storm', 'lib', 'queue', 'list')),
            (True, ('storm', 'pkg', 'get')),
            (True, ('storm', 'impersonate')),
            (True, ('storm', 'service', 'get')),
            (True, ('storm', 'node', 'data')),
        ]

        rules_out_exp = [
            (True, ('foo', 'baz')),
            (True, ('storm',)),
            (True, ('queue', 'get')),
            (True, ('lib', 'queue', 'list')),
            (True, ('pkg', 'get')),
            (True, ('impersonate',)),
            (True, ('service', 'get')),
            (True, ('node', 'data')),
        ]

        rules_out = await migrauth._trnAuthRules(rules_in)
        self.sorteq(rules_out_exp, rules_out)

    async def test_migr_dirn(self):
        with self.getRegrDir('cortexes', REGR_VER) as src:
            with self.getTestDir() as dest:
                os.listdir(os.path.join(src, 'layers'))

                conf = {
                    'src': src,
                    'dest': dest,
                    'addmode': 'nexus',
                }

                async with await s_migr.Migrator.anit(conf) as migr:
                    srcglob = [x for x in glob.glob(src + '/**', recursive=True) if 'lock' not in x]
                    nexpath = os.path.join(dest, 'slabs', 'nexus.lmdb')
                    self.false(os.path.exists(nexpath))

                    await migr._migrDirn()
                    await migr._initStors()
                    self.sorteq(srcglob, [x for x in glob.glob(src + '/**', recursive=True) if 'lock' not in x])
                    self.true(os.path.exists(nexpath))

                    await migr._migrDirn()
                    await migr._initStors()
                    self.sorteq(srcglob, [x for x in glob.glob(src + '/**', recursive=True) if 'lock' not in x])
                    self.true(os.path.exists(nexpath))

    async def test_migr_partiallayer(self):
        podemap = {}
        with self.getRegrDir('assets', REGR_VER) as assetdir:
            podesj = getAssetJson(assetdir, 'podes.json')

            # map by iden and strip out data that wont migrate
            for pode in podesj:
                if pode[0] in NOMIGR_NDEF:
                    continue
                podemap[pode[1]['iden']] = {k: v for k, v in pode[1].get('props', {}).items()}

            podemap = s_common.tuplify(podemap)

        with self.getRegrDir('cortexes', REGR_VER) as src:
            with self.getTestDir() as dest:
                os.listdir(os.path.join(src, 'layers'))

                conf = {
                    'src': src,
                    'dest': dest,
                    'mappartial': True,
                }

                # strip out primary prop rows
                for lyriden in os.listdir(os.path.join(src, 'layers')):

                    lyrdirn = os.path.join(src, 'layers', lyriden, 'layer.lmdb')

                    async with await s_lmdbslab.Slab.anit(lyrdirn) as slab:
                        bybuid = slab.initdb('bybuid')
                        for lkey, _ in slab.scanByFull(db=bybuid):
                            prop = lkey[32:].decode('utf8')
                            if prop[0] == '*':
                                slab.delete(lkey, db=bybuid)

                async with await s_migr.Migrator.anit(conf) as migr:
                    await migr.migrate()

                # startup 2.x core and check layer values
                async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                    layer = core.getLayer()

                    newpodemap = collections.defaultdict(dict)
                    for lkey, lval in layer.layrslab.scanByFull(db='bybuid'):
                        buid = lkey[:32]
                        iden = binascii.hexlify(buid).decode('utf8')

                        flag = lkey[32]
                        self.ne(0, flag)

                        # prop sets only
                        if flag == 1:
                            name = lkey[33:].decode()
                            valu, stortype = s_msgpack.un(lval)

                            newpodemap[iden][name] = valu

                    self.eq(podemap, newpodemap)

    async def test_migr_partiallayer_errors(self):
        with self.getRegrDir('cortexes', REGR_VER) as src:
            with self.getTestDir() as dest:
                os.listdir(os.path.join(src, 'layers'))

                conf = {
                    'src': src,
                    'dest': dest,
                    'mappartial': False,
                }

                # strip out primary prop rows
                for lyriden in os.listdir(os.path.join(src, 'layers')):

                    lyrdirn = os.path.join(src, 'layers', lyriden, 'layer.lmdb')

                    async with await s_lmdbslab.Slab.anit(lyrdirn) as slab:
                        bybuid = slab.initdb('bybuid')
                        for lkey, _ in slab.scanByFull(db=bybuid):
                            prop = lkey[32:].decode('utf8')
                            if prop[0] == '*':
                                slab.delete(lkey, db=bybuid)

                async with await s_migr.Migrator.anit(conf) as migr:
                    # every node will generate an error, but is survivable
                    await migr.migrate()

                # startup 2.x core and check layer values
                async with await s_cortex.Cortex.anit(dest, conf=None) as core:
                    layer = core.getLayer()

                    newpodemap = collections.defaultdict(dict)
                    for lkey, lval in layer.layrslab.scanByFull(db='bybuid'):
                        buid = lkey[:32]
                        iden = binascii.hexlify(buid).decode('utf8')

                        flag = lkey[32]
                        self.ne(0, flag)

                        # prop sets only
                        if flag == 1:
                            name = lkey[33:].decode()
                            valu, stortype = s_msgpack.un(lval)

                            newpodemap[iden][name] = valu

                    self.len(0, newpodemap.keys())

    async def test_migr_migrTriggers(self):
        conf = {
            'src': None,
            'dest': None,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr._migrDirn()
            await migr._initStors()

            # remove useridens
            for iden, valu in migr.cellslab.scanByFull(db=migr.trigdb):
                ruledict = s_msgpack.un(valu)
                ruledict.pop('useriden')
                migr.cellslab.put(iden, s_msgpack.en(ruledict), overwrite=True, db=migr.trigdb)

            await migr._migrTriggers()

            errs = [log async for log in migr._migrlogGet('triggers', 'error')]
            self.isin('Missing iden values for trigger', errs[0].get('val', {}).get('err', ''))

    async def test_migr_trnNodeToNodeedit(self):
        conf = {
            'src': None,
            'dest': None,
        }

        async with self._getTestMigrCore(conf) as (tdata, dest, locallyrs, migr):
            await migr._migrDirn()
            await migr._initStors()
            await migr._migrDatamodel()

            ndef = ('foo:bar', '1.2.3.3')
            buid = s_common.buid(ndef)
            node = await migr._srcPackNode(buid, ndef, {}, {}, {})
            err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
            self.none(ne)
            self.isin('Unable to parse form name', err['mesg'])

            ndef = ('*inet:fqdn', 123)
            buid = s_common.buid(ndef)
            node = await migr._srcPackNode(buid, ndef, {}, {}, {})
            err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
            self.none(ne)
            self.isin('Buid/norming exception', err['mesg'])

            ndef = ('*inet:fqdn', 'foo.com.')
            buid = s_common.buid(('inet:fqdn', 'foo.com.'))
            node = await migr._srcPackNode(buid, ndef, {}, {}, {})
            err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
            self.none(ne)
            self.isin('Normed form val does not match inbound', err['mesg'])

            ndef = ('*inet:fqdn', 'foo.com.')
            buid = s_common.buid(('inet:fqdn', 'foo.com'))
            node = await migr._srcPackNode(buid, ndef, {}, {}, {})
            err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
            self.none(ne)
            self.isin('Calculated buid does not match inbound', err['mesg'])

            ndef = ('*inet:fqdn', 'foo.com')
            buid = s_common.buid(('inet:fqdn', 'foo.com'))
            node = await migr._srcPackNode(buid, ndef, {'foo:bar': 'ham'}, {}, {})
            err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
            self.none(ne)
            self.isin('Unable to determine stortype for sprop foo:bar', err['mesg'])

            ndef = ('*inet:fqdn', 'foo.com')
            buid = s_common.buid(('inet:fqdn', 'foo.com'))
            node = await migr._srcPackNode(buid, ndef, {'inet:ipv4': 'ham'}, {}, {})
            err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
            self.none(ne)
            self.isin('Unable to determine stortype for sprop inet:ipv4', err['mesg'])

            ndef = ('*inet:fqdn', 'foo.com')
            buid = s_common.buid(('inet:fqdn', 'foo.com'))
            node = await migr._srcPackNode(buid, ndef, {}, {}, {'newp': {'ahh': 37}})
            err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
            self.none(ne)
            self.isin('Unable to determine stortype for tagprop ahh', err['mesg'])

            lyrinfo = await migr._migrHiveLayerInfo(locallyrs[0])
            async with migr._destGetWlyr(migr.dest, locallyrs[0], lyrinfo) as wlyr:

                # continue on bad nodeedits
                ndef = ('*inet:fqdn', 'foo.com')
                buid = s_common.buid(('inet:fqdn', 'foo.com'))
                node = await migr._srcPackNode(buid, ndef, {}, {}, {})
                err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
                self.none(err)
                ne = ('badbuid', 'foo:bar', ne[2])

                res = await migr._destAddNodes(wlyr, [ne], 'nexus')
                self.isin('Unable to store nodeedits', res.get('mesg', ''))

                # test array types
                mdef = {
                    'types': (
                        ('test:array', ('array', {'type': 'inet:ipv4'}), {}),
                    ),
                    'forms': (
                        ('test:array', {}, (
                        )),
                    ),
                }
                migr.model.addDataModels([('asdf', mdef)])

                ndef = ('*test:array', (16909060, 84281096))
                buid = s_common.buid(('test:array', (16909060, 84281096)))
                node = await migr._srcPackNode(buid, ndef, {}, {}, {})
                err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
                res = await migr._destAddNodes(wlyr, [ne], 'editor')
                self.none(res)

                ndef = ('*ou:org', '4386f6146f3eb3be784d3e0d44cede5e')
                buid = s_common.buid(('ou:org', '4386f6146f3eb3be784d3e0d44cede5e'))
                node = await migr._srcPackNode(buid, ndef, {'names': ('foo', 'bar')}, {}, {})
                err, ne = await migr._trnNodeToNodeedit(node, migr.model, fname=None, chknodes=True)
                res = await migr._destAddNodes(wlyr, [ne], 'editor')
                self.none(res)

                # test inet:server:proto=host
                # original valu was host://fazbaz.com
                ndef = ('*inet:server', 'host://1144e562c90bf6d2bbb3a80b1982b84f')
                buid = s_common.buid(('inet:server', 'host://1144e562c90bf6d2bbb3a80b1982b84f'))
                props = {'proto': 'host', 'host': '1144e562c90bf6d2bbb3a80b1982b84f', '.created': 1586437148551}
                node = await migr._srcPackNode(buid, ndef, props, {}, {})

                # no safety check
                err, ne = await migr._trnNodeToNodeedit(node, migr.model, chknodes=False)
                self.none(err)
                sodes = await wlyr.storNodeEdits([ne], meta=None)  # returns sode
                self.len(1, sodes)
                self.eq(buid, sodes[0][0])

                # safety check on to exercise re-norming bypass
                err, ne = await migr._trnNodeToNodeedit(node, migr.model, chknodes=True)
                self.none(err)
                sodes = await wlyr.storNodeEdits([ne], meta=None)  # returns sode
                self.len(1, sodes)
                self.eq(buid, sodes[0][0])



import synapse.exc as s_exc
import synapse.common as s_common

from synapse.common import aspin

import synapse.telepath as s_telepath
import synapse.tests.utils as s_t_utils

class TrigTest(s_t_utils.SynTest):

    async def test_trigger_recursion(self):
        async with self.getTestCore() as core:
            tdef = {'cond': 'node:add', 'form': 'test:guid', 'storm': '[ test:guid="*" ]'}
            await core.view.addTrigger(tdef)
            await self.asyncraises(s_exc.RecursionLimitHit, core.nodes('[ test:guid="*" ]'))

    async def test_modification_persistence(self):

        with self.getTestDir() as fdir:

            async with self.getTestCore(dirn=fdir) as core:
                iden = s_common.guid()
                tdef = {'cond': 'node:add', 'form': 'inet:ipv4', 'storm': '[inet:user=1] | testcmd'}
                await core.view.addTrigger(tdef)

                triggers = await core.view.listTriggers()
                self.eq(triggers[0][1].tdef['storm'], '[inet:user=1] | testcmd')

                iden = triggers[0][0]
                await core.view.setTriggerInfo(iden, 'storm', '[inet:user=2 .test:univ=4] | testcmd')
                triggers = await core.view.listTriggers()
                self.eq(triggers[0][1].tdef['storm'], '[inet:user=2 .test:univ=4] | testcmd')

                # Sad cases
                await self.asyncraises(s_exc.BadSyntax, core.view.setTriggerInfo(iden, 'storm', ' | | badstorm '))
                await self.asyncraises(s_exc.NoSuchIden, core.view.setTriggerInfo('deadb33f', 'storm', 'inet:user'))

    async def test_trigger_basics(self):

        async with self.getTestCoreAndProxy() as (core, prox):

            view = core.view

            # node:add case
            tdef = {'cond': 'node:add', 'form': 'test:str', 'storm': '[ test:guid="*" +#nodeadd]'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo ]')
            self.len(1, await core.nodes('test:guid#nodeadd'))

            # node:del case
            tdef = {'cond': 'node:del', 'storm': '[ test:guid="*" +#nodedel ]', 'form': 'test:str'}
            await view.addTrigger(tdef)
            await core.nodes('test:str=foo | delnode')
            self.len(1, await core.nodes('test:guid#nodedel'))

            # tag:add case
            tdef = {'cond': 'tag:add', 'storm': '[ test:guid="*" +#tagadd]', 'tag': 'footag'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#footag ]')
            self.len(1, await core.nodes('test:guid#tagadd'))

            # tag:add globbing and storm var
            tdef = {'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]', 'tag': 'a.*.c'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#a.b ]')
            await core.nodes('[ test:str=foo +#a.b.c ]')
            await core.nodes('[ test:str=foo +#a.b.ccc ]')
            self.len(1, await core.nodes('#count'))
            self.len(1, await core.nodes('test:str=a.b.c'))

            tdef = {'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]', 'tag': 'foo.**.baz'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#foo.1.2.3.baz ]')
            self.len(1, await core.nodes('test:str=foo.1.2.3.baz'))

            # tag:del case
            tdef = {'cond': 'tag:del', 'storm': '[ test:guid="*" +#tagdel ]', 'tag': 'footag'}
            await view.addTrigger(tdef)
            await core.nodes('test:str=foo [ -#footag ]')
            self.len(1, await core.nodes('test:guid#tagdel'))

            # Form/tag add
            tdef = {'cond': 'tag:add', 'storm': '[ test:guid="*" +#formtagadd]', 'tag': 'bartag', 'form': 'test:str'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#bartag ]')
            self.len(1, await core.nodes('test:guid#formtagadd'))

            # Wrong form/right tag add doesn't fire
            await core.nodes('test:int=5 | delnode')
            self.len(0, await core.nodes('test:int=5'))
            await core.nodes('[ test:auto=1 +#bartag ]')
            self.len(0, await core.nodes('test:int=5'))

            # Right form/wrong tag add doesn't fire
            await core.nodes('[ test:str=bar +#footag ]')
            self.len(0, await core.nodes('test:int=5'))

            # Prop set
            tdef = {'cond': 'prop:set', 'storm': '[ test:guid="*" +#propset ]', 'prop': 'test:type10:intprop'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:type10=1 ]')
            await core.nodes('[ test:type10=1 :intprop=25 ]')
            self.len(1, await core.nodes('test:guid#propset'))

            # Test re-setting doesn't fire
            await core.nodes('test:int=6 | delnode')
            await core.nodes('[ test:type10=1 :intprop=25 ]')
            self.len(0, await core.nodes('test:int=6'))

            # Prop set univ
            tdef = {'cond': 'prop:set', 'storm': '[ test:guid="*" +#propsetuniv ]', 'prop': '.test:univ'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:type10=1 .test:univ=1 ]')
            self.len(1, await core.nodes('test:guid#propsetuniv'))

            # Prop set form specific univ
            tdef = {'cond': 'prop:set', 'storm': '[ test:guid="*" +#propsetuniv2 ]', 'prop': 'test:str.test:univ'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=beep .test:univ=1 ]')
            self.len(1, await core.nodes('test:guid#propsetuniv2'))

            # Add trigger with iden
            iden = s_common.guid()
            tdef0 = {'cond': 'node:add', 'storm': '[ +#withiden ]', 'form': 'test:int', 'iden': iden}
            await view.addTrigger(tdef0)
            self.nn(await view.getTrigger(iden))
            await core.nodes('[ test:int=77 ]')
            self.len(1, await core.nodes('test:int#withiden'))

            # Attempting to add trigger with existing iden raises
            with self.raises(s_exc.DupIden):
                tdef = {'cond': 'node:add', 'storm': '[ +#dupiden ]', 'form': 'test:int', 'iden': iden}
                await view.addTrigger(tdef)
            self.eq(tdef0, (await view.getTrigger(iden)).tdef)

            # Bad trigger parms
            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'nocond', 'storm': 'test:int=4', 'form': 'test:str'})

            with self.raises(s_exc.BadSyntax):
                await view.addTrigger({'cond': 'node:add', 'storm': ' | | badstorm ', 'form': 'test:str'})

            with self.raises(s_exc.BadOptValu):
                await view.addTrigger({'cond': 'node:add', 'storm': 'test:int=4', 'form': 'test:str', 'tag': 'foo'})

            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'prop:set', 'storm': 'test:int=4', 'form': 'test:str', 'prop': 'foo'})

            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]'})

            with self.raises(s_exc.BadOptValu):
                tdef = {'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]', 'tag': 'foo', 'prop': 'test:str'}
                await view.addTrigger(tdef)

            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'node:add', 'storm': 'test:int=4', 'form': 'test:str', 'iden': 'foo'})

            # bad tagmatch
            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]', 'tag': 'foo&baz'})

            # Trigger list
            triglist = await view.listTriggers()
            self.len(11, triglist)

            # Delete not a trigger
            await self.asyncraises(s_exc.NoSuchIden, view.delTrigger('foo'))

            # Delete trigger
            iden = [iden for iden, r in triglist if r.tdef['cond'] == 'prop:set'][0]
            await core.nodes('test:int=6 | delnode')
            await view.delTrigger(iden)

            # The deleted trigger shall not fire
            await core.nodes('[ test:type10=3 :intprop=25 ]')
            self.len(0, await core.nodes('test:int=6'))

            await self.asyncraises(s_exc.NoSuchIden, view.delTrigger(b'badiden'))

            # Mod trigger

            iden = [iden for iden, r in triglist if r.tdef['cond'] == 'tag:add' and r.tdef.get('form') == 'test:str'][0]

            await view.setTriggerInfo(iden, 'storm', '[ test:int=42 ]')
            await core.nodes('[ test:str=foo4 +#bartag ]')
            self.len(1, await core.nodes('test:int=42'))

            # Delete a tag:add
            iden2 = [iden for iden, r in triglist if r.tdef['cond'] == 'tag:add' and r.tdef.get('form') is None][0]
            await view.delTrigger(iden2)

            # A rando user can't manipulate triggers

            user = await core.auth.addUser('fred')
            await user.setPasswd('fred')

            url = core.getLocalUrl()

            async with await s_telepath.openurl(url, user='fred') as fred:
                # Trigger list other user
                self.len(0, await fred.callStorm('return ($lib.trigger.list())'))

                with self.raises(s_exc.AuthDeny):
                    tdef = {'cond': 'node:add', 'storm': '[ test:int=1 ]', 'form': 'test:str'}
                    opts = {'vars': {'tdef': tdef}}
                    q = 'return ($lib.trigger.add($tdef))'
                    await fred.callStorm(q, opts=opts)

                # Delete trigger auth failure
                await self.asyncraises(s_exc.StormRuntimeError, fred.callStorm(f'$lib.trigger.del({iden})'))

                # Mod trigger auth failure
                opts = {'vars': {'iden': iden}}
                await self.asyncraises(s_exc.StormRuntimeError,
                                       fred.callStorm('$lib.trigger.mod($iden, "{#foo}")', opts=opts))

            # additional NoSuchIden failures
            await self.asyncraises(s_exc.NoSuchIden, view.getTrigger('newp'))
            await self.asyncraises(s_exc.NoSuchIden, view.delTrigger('newp'))
            await self.asyncraises(s_exc.NoSuchIden, view.setTriggerInfo('newp', 'enabled', True))

    async def test_trigger_delete(self):

        async with self.getTestCore() as core:

            tdef0 = {'cond': 'node:add', 'form': 'test:guid', 'storm': '[test:str=add]'}
            tdef1 = {'cond': 'node:del', 'form': 'test:guid', 'storm': '[test:str=del]'}
            tdef2 = {'cond': 'prop:set', 'prop': 'test:guid:tick', 'storm': '[test:str=set]'}

            tdef0 = await core.view.addTrigger(tdef0)
            tdef1 = await core.view.addTrigger(tdef1)
            tdef2 = await core.view.addTrigger(tdef2)

            await core.nodes('[test:guid="*" :tick=2015] | delnode')
            self.len(3, await core.nodes('test:str'))

            await core.view.delTrigger(tdef0.get('iden'))
            await core.view.delTrigger(tdef1.get('iden'))
            await core.view.delTrigger(tdef2.get('iden'))

            await core.nodes('test:str | delnode')
            await core.nodes('[test:guid="*" :tick=2015] | delnode')

            self.len(0, await core.nodes('test:str'))

    async def test_trigger_tag_globs(self):

        async with self.getTestCore() as core:

            root = await core.auth.getUserByName('root')

            tdef = {'iden': '1', 'user': root.iden, 'cond': 'tag:add', 'storm': '$lib.queue.get(foo).put(count0)',
                    'tag': 'foo.*.bar', 'enabled': True}
            trig1 = core.view.triggers.load(tdef)

            tdef = {'iden': '2', 'user': root.iden, 'cond': 'tag:del', 'storm': '$lib.queue.get(foo).put(count1)',
                    'tag': 'baz.*.faz', 'form': 'test:guid', 'enabled': True}
            trig2 = core.view.triggers.load(tdef)

            await core.nodes('$lib.queue.add(foo)')

            async def popNextFoo():
                return await core.callStorm('''
                    return ($lib.queue.get(foo).pop().index(1))
                ''')

            await core.nodes('[ test:guid="*" +#foo.asdf.bar ]')
            await core.nodes('[ test:guid="*" +#baz.asdf.faz ]')
            await core.nodes('#baz.asdf.faz [ -#baz.asdf.faz ]')

            self.eq('count0', await popNextFoo())
            self.eq('count1', await popNextFoo())

            core.view.triggers.pop(trig1.iden)
            core.view.triggers.pop(trig2.iden)

            await core.nodes('test:guid | delnode')

            # Triggers don't fire after they've been deleted

            await core.nodes('[ test:guid="*" +#foo.asdf.bar ]')
            await core.nodes('[ test:guid="*" +#baz.asdf.faz ]')

            await core.nodes('#baz.asdf.faz [ -#baz.asdf.faz ]')

            self.eq(0, await core.callStorm('return ($lib.queue.get(foo).size())'))

    async def test_trigger_perms(self):

        async with self.getTestCore() as core:

            visi = await core.auth.addUser('visi')
            newb = await core.auth.addUser('newb')

            await visi.addRule((True, ('node', 'tag', 'add', 'foo')))

            async with core.getLocalProxy(user='visi') as proxy:

                with self.raises(s_exc.AuthDeny):
                    tdef = {'cond': 'node:add', 'form': 'inet:ipv4', 'storm': '[ +#foo ]'}
                    await proxy.callStorm('return ($lib.trigger.add($tdef).get(iden))', opts={'vars': {'tdef': tdef}})

                await visi.addRule((True, ('trigger', 'add')))

                tdef = {'cond': 'node:add', 'form': 'inet:ipv4', 'storm': '[ +#foo ]'}
                opts = {'vars': {'tdef': tdef}}
                trig = await proxy.callStorm('return ($lib.trigger.add($tdef))', opts=opts)
                iden0 = trig['iden']

                iden1 = s_common.guid()
                tdef = {'cond': 'node:add', 'form': 'inet:ipv6', 'storm': '[ +#foo ]', 'iden': iden1}
                opts = {'vars': {'tdef': tdef}}
                trig = await proxy.callStorm('return ($lib.trigger.add($tdef))', opts=opts)

                nodes = await core.nodes('[ inet:ipv4=1.2.3.4 ]')
                self.nn(nodes[0].tags.get('foo'))

                await aspin(proxy.eval('$lib.trigger.del($iden)', opts={'vars': {'iden': iden1}}))

            trigs = await core.view.listTriggers()
            trigiden = trigs[0][0]
            self.eq(trigiden, iden0)

            async with core.getLocalProxy(user='newb') as proxy:

                await self.agenlen(1, proxy.eval('syn:trigger'))

                await newb.addRule((True, ('trigger', 'get')))
                with self.raises(s_exc.AuthDeny):
                    await proxy.eval('$lib.trigger.del($iden)', opts={'vars': {'iden': trigs[0][0]}}).list()

                await self.agenlen(1, proxy.eval('syn:trigger'))

                with self.raises(s_exc.AuthDeny):
                    opts = {'vars': {'iden': trigiden}}
                    await proxy.eval('$lib.trigger.get($iden).set(enabled, $(0))', opts=opts).list()

                await newb.addRule((True, ('trigger', 'set')))
                opts = {'vars': {'iden': trigiden}}
                await aspin(proxy.eval('$lib.trigger.get($iden).set(enabled, $(0))', opts=opts))

                await newb.addRule((True, ('trigger', 'del')))
                await aspin(proxy.eval('$lib.trigger.del($iden)', opts={'vars': {'iden': trigiden}}))

    async def test_trigger_runts(self):

        async with self.getTestCore() as core:

            tdef = await core.view.addTrigger({
                'cond': 'node:add',
                'form': 'test:str',
                'storm': '[ test:int=1 ]',
            })
            iden = tdef['iden']

            nodes = await core.nodes('syn:trigger')
            self.len(1, nodes)
            self.eq(nodes[0].get('doc'), '')

            nodes = await core.nodes(f'syn:trigger={iden} [ :doc="hehe haha" :name=visitrig ]')
            self.eq(nodes[0].get('doc'), 'hehe haha')
            self.eq(nodes[0].get('name'), 'visitrig')

            nodes = await core.nodes(f'syn:trigger={iden}')
            self.eq(nodes[0].get('doc'), 'hehe haha')
            self.eq(nodes[0].get('name'), 'visitrig')



import synapse.lib.scrape as s_scrape

import synapse.tests.utils as s_t_utils

data0 = '''

visi@vertex.link is an email address

and BOB@WOOT.COM is another

    hehe.taxi

    aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa

    aa:bb:cc:dd:ee:ff

    1.2.3.4

    5.6.7.8:16

'''

class ScrapeTest(s_t_utils.SynTest):

    def test_scrape(self):

        nodes = set(s_scrape.scrape(data0))

        self.len(9, nodes)
        nodes.remove(('hash:md5', 'a' * 32))
        nodes.remove(('inet:ipv4', '1.2.3.4'))
        nodes.remove(('inet:ipv4', '5.6.7.8'))
        nodes.remove(('inet:fqdn', 'WOOT.COM'))
        nodes.remove(('inet:fqdn', 'hehe.taxi'))
        nodes.remove(('inet:fqdn', 'vertex.link'))
        nodes.remove(('inet:server', '5.6.7.8:16'))
        nodes.remove(('inet:email', 'BOB@WOOT.COM'))
        nodes.remove(('inet:email', 'visi@vertex.link'))
        self.len(0, nodes)

        nodes = set(s_scrape.scrape(data0, 'inet:email'))
        self.len(2, nodes)
        nodes.remove(('inet:email', 'BOB@WOOT.COM'))
        nodes.remove(('inet:email', 'visi@vertex.link'))
        self.len(0, nodes)

    def test_scrape_sequential(self):
        md5 = ('a' * 32, 'b' * 32, )
        sha1 = ('c' * 40, 'd' * 40, )
        sha256 = ('e' * 64, 'f' * 64, )
        url = ('http://foobar.com', 'http://cat.net', )
        ipv4 = ('1.2.3.4', '5.6.7.8', )
        server = ('7.7.7.7:123', '8.8.8.8:456', )
        fqdn = ('woot.com', 'baz.io', )
        email = ('me@bar.io', 'you@zee.com', )

        txt = f'hehe {md5[0]} {md5[1]} haha'
        self.eq({md5[0], md5[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {md5[0]},{md5[1]} haha'
        self.eq({md5[0], md5[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {sha1[0]} {sha1[1]} haha'
        self.eq({sha1[0], sha1[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {sha256[0]} {sha256[1]} haha'
        self.eq({sha256[0], sha256[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {url[0]} {url[1]} haha'
        self.eq({url[0], 'foobar.com', url[1], 'cat.net', }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {ipv4[0]} {ipv4[1]} haha'
        self.eq({ipv4[0], ipv4[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {server[0]} {server[1]} haha'
        self.eq({server[0], '7.7.7.7', server[1], '8.8.8.8', }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe "{fqdn[0]}" "{fqdn[1]}" haha'
        self.eq({fqdn[0], fqdn[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {fqdn[0]}  {fqdn[1]} haha'
        self.eq({fqdn[0], fqdn[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {email[0]}, {email[1]} haha'
        self.eq({email[0], 'bar.io', email[1], 'zee.com', }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {fqdn[0]}. {fqdn[1]} haha'
        self.eq({fqdn[0], fqdn[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {fqdn[0]},{fqdn[1]} haha'
        self.eq({fqdn[0], fqdn[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {fqdn[0]} {fqdn[1]} haha'
        self.eq({fqdn[0], fqdn[1], }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {email[0]}. {email[1]} haha'
        self.eq({email[0], 'bar.io', email[1], 'zee.com', }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {email[0]} {email[1]} haha'
        self.eq({email[0], 'bar.io', email[1], 'zee.com', }, {n[1] for n in s_scrape.scrape(txt)})

        txt = f'hehe {email[0]} {fqdn[0]} haha'
        self.eq({email[0], 'bar.io', fqdn[0], }, {n[1] for n in s_scrape.scrape(txt)})



from __future__ import annotations

import types
import asyncio
import logging
import weakref
import contextlib
import collections

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.coro as s_coro
import synapse.lib.base as s_base
import synapse.lib.node as s_node
import synapse.lib.cache as s_cache
import synapse.lib.layer as s_layer
import synapse.lib.storm as s_storm
import synapse.lib.types as s_types
import synapse.lib.spooled as s_spooled

logger = logging.getLogger(__name__)

class Snap(s_base.Base):
    '''
    A "snapshot" is a transaction across multiple Cortex layers.

    The Snap object contains the bulk of the Cortex API to
    facilitate performance through careful use of transaction
    boundaries.

    Transactions produce the following EventBus events:

    (...any splice...)
    ('log', {'level': 'mesg': })
    ('print', {}),
    '''

    async def __anit__(self, view, user):
        '''
        Args:
            core (cortex):  the cortex
            layers (List[Layer]): the list of layers to access, write layer last
        '''
        await s_base.Base.__anit__(self)

        self.stack = contextlib.ExitStack()

        assert user is not None

        self.strict = True
        self.elevated = False
        self.canceled = False

        self.core = view.core
        self.view = view
        self.user = user

        self.buidprefetch = self.view.isafork()

        self.layers = list(reversed(view.layers))
        self.wlyr = self.layers[-1]

        self.readonly = self.wlyr.readonly

        # variables used by the storm runtime
        self.vars = {}

        self.runt = {}

        self.debug = False      # Set to true to enable debug output.
        self.write = False      # True when the snap has a write lock on a layer.

        self._tagcachesize = 10000
        self._buidcachesize = 100000
        self.tagcache = s_cache.FixedCache(self._addTagNode, size=self._tagcachesize)
        # Keeps alive the most recently accessed node objects
        self.buidcache = collections.deque(maxlen=self._buidcachesize)
        self.livenodes = weakref.WeakValueDictionary()  # buid -> Node

        self.onfini(self.stack.close)
        self.changelog = []
        self.tagtype = self.core.model.type('ival')
        self.trigson = self.core.trigson

    def disableTriggers(self):
        self.trigson = False

    async def getSnapMeta(self):
        '''
        Retrieve snap metadata to store along side nodeEdits.
        '''
        meta = {
            'time': s_common.now(),
            'user': self.user.iden
        }

        providen = self.core.provstor.precommit()
        if providen is not None:
            meta['prov'] = providen

        return meta

    @contextlib.contextmanager
    def getStormRuntime(self, opts=None, user=None):
        if user is None:
            user = self.user

        runt = s_storm.Runtime(self, opts=opts, user=user)
        runt.isModuleRunt = True

        yield runt

    async def iterStormPodes(self, text, opts=None, user=None):
        '''
        Yield packed node tuples for the given storm query text.
        '''
        if user is None:
            user = self.user

        dorepr = False
        dopath = False

        self.core._logStormQuery(text, user)

        if opts is not None:
            dorepr = opts.get('repr', False)
            dopath = opts.get('path', False)

        async for node, path in self.storm(text, opts=opts, user=user):
            pode = node.pack(dorepr=dorepr)
            pode[1]['path'] = path.pack(path=dopath)
            yield pode

    @s_coro.genrhelp
    async def storm(self, text, opts=None, user=None):
        '''
        Execute a storm query and yield (Node(), Path()) tuples.
        '''
        if user is None:
            user = self.user

        if opts is None:
            opts = {}

        mode = opts.get('mode', 'storm')

        query = self.core.getStormQuery(text, mode=mode)
        with self.getStormRuntime(opts=opts, user=user) as runt:
            async for x in runt.iterStormQuery(query):
                yield x

    @s_coro.genrhelp
    async def eval(self, text, opts=None, user=None):
        '''
        Run a storm query and yield Node() objects.
        '''
        if user is None:
            user = self.user

        if opts is None:
            opts = {}

        mode = opts.get('mode', 'storm')

        # maintained for backward compatibility
        query = self.core.getStormQuery(text, mode=mode)
        with self.getStormRuntime(opts=opts, user=user) as runt:
            async for node, path in runt.iterStormQuery(query):
                yield node

    async def nodes(self, text, opts=None, user=None):
        return [n async for n in self.eval(text, opts=opts, user=user)]

    async def clearCache(self):
        self.tagcache.clear()
        self.buidcache.clear()
        self.livenodes.clear()

    async def printf(self, mesg):
        await self.fire('print', mesg=mesg)

    async def warn(self, mesg, log=True, **info):
        if log:
            logger.warning(mesg)
        await self.fire('warn', mesg=mesg, **info)

    async def getNodeByBuid(self, buid):
        '''
        Retrieve a node tuple by binary id.

        Args:
            buid (bytes): The binary ID for the node.

        Returns:
            Optional[s_node.Node]: The node object or None.

        '''
        return await self._joinStorNode(buid, {})

    async def getNodeByNdef(self, ndef):
        '''
        Return a single Node by (form,valu) tuple.

        Args:
            ndef ((str,obj)): A (form,valu) ndef tuple.  valu must be
            normalized.

        Returns:
            (synapse.lib.node.Node): The Node or None.
        '''
        buid = s_common.buid(ndef)
        return await self.getNodeByBuid(buid)

    async def nodesByTagProp(self, form, tag, name):
        prop = self.core.model.getTagProp(name)
        if prop is None:
            mesg = f'No tag property named {name}'
            raise s_exc.NoSuchTagProp(name=name, mesg=mesg)

        for layr in self.layers:
            genr = layr.liftByTagProp(form, tag, name)
            async for node in self._joinStorGenr(layr, genr):
                yield node

    async def nodesByTagPropValu(self, form, tag, name, cmpr, valu):

        prop = self.core.model.getTagProp(name)
        if prop is None:
            mesg = f'No tag property named {name}'
            raise s_exc.NoSuchTagProp(name=name, mesg=mesg)

        cmprvals = prop.type.getStorCmprs(cmpr, valu)
        # an empty return probably means ?= with invalid value
        if not cmprvals:
            return

        for layr in self.layers:
            genr = layr.liftByTagPropValu(form, tag, name, cmprvals)
            async for node in self._joinStorGenr(layr, genr):
                if node.bylayer['tagprops'].get((tag, prop.name)) != layr:
                    continue
                yield node

    async def _joinStorNode(self, buid, cache):

        node = self.livenodes.get(buid)
        if node is not None:
            await asyncio.sleep(0)
            return node

        ndef = None

        tags = {}
        props = {}
        nodedata = {}
        tagprops = {}

        bylayer = {
            'ndef': None,
            'tags': {},
            'props': {},
            'tagprops': {},
        }

        for layr in self.layers:

            sode = cache.get(layr.iden)
            if sode is None:
                sode = await layr.getStorNode(buid)

            info = sode[1]

            storndef = info.get('ndef')
            if storndef is not None:
                ndef = storndef
                bylayer['ndef'] = layr

            storprops = info.get('props')
            if storprops is not None:
                props.update(storprops)
                bylayer['props'].update({p: layr for p in storprops.keys()})

            stortags = info.get('tags')
            if stortags is not None:
                tags.update(stortags)
                bylayer['tags'].update({p: layr for p in stortags.keys()})

            stortagprops = info.get('tagprops')
            if stortagprops is not None:
                tagprops.update(stortagprops)
                bylayer['tagprops'].update({p: layr for p in stortagprops.keys()})

            stordata = info.get('nodedata')
            if stordata is not None:
                nodedata.update(stordata)

        if ndef is None:
            return None

        fullnode = (buid, {
            'ndef': ndef,
            'tags': tags,
            'props': props,
            'nodedata': nodedata,
            'tagprops': tagprops,
        })

        node = s_node.Node(self, fullnode, bylayer=bylayer)
        self.livenodes[buid] = node
        self.buidcache.append(node)

        # moved here from getNodeByBuid() to cover more
        await asyncio.sleep(0)
        return node

    async def _joinStorGenr(self, layr, genr):
        cache = {}
        async for sode in genr:
            cache[layr.iden] = sode
            node = await self._joinStorNode(sode[0], cache)
            if node is not None:
                yield node

    async def nodesByDataName(self, name):
        for layr in self.layers:
            genr = layr.liftByDataName(name)
            async for node in self._joinStorGenr(layr, genr):
                yield node

    async def nodesByProp(self, full):

        prop = self.core.model.prop(full)
        if prop is None:
            mesg = f'No property named "{full}".'
            raise s_exc.NoSuchProp(mesg=mesg)

        if prop.isrunt:

            async for node in self.getRuntNodes(prop.full):
                yield node

            return

        if prop.isform:

            for layr in self.layers:
                genr = layr.liftByProp(prop.name, None)

                async for node in self._joinStorGenr(layr, genr):

                    # TODO merge sort rather than use bylayer
                    if node.bylayer.get('ndef') != layr:
                        continue

                    yield node
            return

        if prop.isuniv:

            for layr in self.layers:
                genr = layr.liftByProp(None, prop.name)
                async for node in self._joinStorGenr(layr, genr):
                    if node.bylayer['props'].get(prop.name) != layr:
                        continue
                    yield node
            return

        formname = None
        if not prop.isuniv:
            formname = prop.form.name

        # Prop is secondary prop

        for layr in self.layers:
            genr = layr.liftByProp(formname, prop.name)
            async for node in self._joinStorGenr(layr, genr):
                if node.bylayer['props'].get(prop.name) != layr:
                    continue
                yield node

    async def nodesByPropValu(self, full, cmpr, valu):

        if cmpr == 'type=':
            async for node in self.nodesByPropValu(full, '=', valu):
                yield node

            async for node in self.nodesByPropTypeValu(full, valu):
                yield node
            return

        prop = self.core.model.prop(full)
        if prop is None:
            mesg = f'No property named "{full}".'
            raise s_exc.NoSuchProp(mesg=mesg)

        cmprvals = prop.type.getStorCmprs(cmpr, valu)
        # an empty return probably means ?= with invalid value
        if not cmprvals:
            return

        if prop.isrunt:
            for storcmpr, storvalu, _ in cmprvals:
                async for node in self.getRuntNodes(prop.full, valu=storvalu, cmpr=storcmpr):
                    yield node
            return

        if prop.isform:

            for layr in self.layers:
                genr = layr.liftByFormValu(prop.name, cmprvals)

                async for node in self._joinStorGenr(layr, genr):

                    # TODO merge sort rather than use bylayer
                    if node.bylayer.get('ndef') != layr:
                        continue

                    yield node

            return

        if prop.isuniv:

            for layr in self.layers:
                genr = layr.liftByPropValu(None, prop.name, cmprvals)
                async for node in self._joinStorGenr(layr, genr):
                    if node.bylayer['props'].get(prop.name) != layr:
                        continue
                    yield node

            return

        for layr in self.layers:
            genr = layr.liftByPropValu(prop.form.name, prop.name, cmprvals)
            async for node in self._joinStorGenr(layr, genr):
                if node.bylayer['props'].get(prop.name) != layr:
                    continue
                yield node

    async def nodesByTag(self, tag, form=None):
        for layr in self.layers:
            genr = layr.liftByTag(tag, form=form)
            async for node in self._joinStorGenr(layr, genr):
                if node.bylayer['tags'].get(tag) != layr:
                    continue
                yield node

    async def nodesByTagValu(self, tag, cmpr, valu, form=None):
        norm, info = self.core.model.type('ival').norm(valu)
        for layr in self.layers:
            genr = layr.liftByTagValu(tag, cmpr, norm, form=form)
            async for node in self._joinStorGenr(layr, genr):
                if node.bylayer['tags'].get(tag) != layr:
                    continue
                yield node

    async def nodesByPropTypeValu(self, name, valu):

        _type = self.core.model.types.get(name)
        if _type is None:
            raise s_exc.NoSuchType(name=name)

        for prop in self.core.model.getPropsByType(name):
            async for node in self.nodesByPropValu(prop.full, '=', valu):
                yield node

    async def nodesByPropArray(self, full, cmpr, valu):

        prop = self.core.model.prop(full)
        if prop is None:
            mesg = f'No property named "{full}".'
            raise s_exc.NoSuchProp(mesg=mesg)

        if not isinstance(prop.type, s_types.Array):
            mesg = f'Array synax is invalid on non array type: {prop.type.name}.'
            raise s_exc.BadTypeValu(mesg=mesg)

        cmprvals = prop.type.arraytype.getStorCmprs(cmpr, valu)

        if prop.isform:

            for layr in self.layers:
                genr = layr.liftByPropArray(prop.name, None, cmprvals)
                async for node in self._joinStorGenr(layr, genr):
                    if node.bylayer['ndef'] != layr:
                        continue
                    yield node

            return

        formname = None
        if prop.form is not None:
            formname = prop.form.name

        for layr in self.layers:
            genr = layr.liftByPropArray(formname, prop.name, cmprvals)
            async for node in self._joinStorGenr(layr, genr):
                if node.bylayer['props'].get(prop.name) != layr:
                    continue
                yield node

    async def getNodeAdds(self, form, valu, props, addnode=True):

        async def _getadds(f, p, formnorm, forminfo, doaddnode=True):

            edits = []  # Non-primary prop edits
            topsubedits = []  # Primary prop sub edits

            formsubs = forminfo.get('subs', {})
            for subname, subvalu in formsubs.items():
                p[subname] = subvalu

            for propname, propvalu in p.items():
                subedits: s_layer.NodeEditsT = []

                prop = f.prop(propname)
                if prop is None:
                    continue

                assert prop.type.stortype is not None

                if isinstance(prop.type, s_types.Ndef):
                    ndefname, ndefvalu = propvalu
                    ndefform = self.core.model.form(ndefname)
                    if ndefform is None:
                        raise s_exc.NoSuchForm(name=ndefname)

                    ndefnorm, ndefinfo = ndefform.type.norm(ndefvalu)
                    do_subedit = True
                    if self.buidprefetch:
                        node = await self.getNodeByBuid(s_common.buid((ndefform.name, ndefnorm)))
                        do_subedit = node is None

                    if do_subedit:
                        subedits.extend([x async for x in _getadds(ndefform, {}, ndefnorm, ndefinfo)])

                elif isinstance(prop.type, s_types.Array):
                    arrayform = self.core.model.form(prop.type.arraytype.name)
                    if arrayform is not None:
                        for arrayvalu in propvalu:
                            arraynorm, arrayinfo = arrayform.type.norm(arrayvalu)

                            if self.buidprefetch:
                                node = await self.getNodeByBuid(s_common.buid((arrayform.name, arraynorm)))
                                if node is not None:
                                    continue

                            subedits.extend([x async for x in _getadds(arrayform, {}, arraynorm, arrayinfo)])

                propnorm, typeinfo = prop.type.norm(propvalu)

                propsubs = typeinfo.get('subs')
                if propsubs is not None:
                    for subname, subvalu in propsubs.items():
                        fullname = f'{prop.full}:{subname}'
                        subprop = self.core.model.prop(fullname)
                        if subprop is None:
                            continue

                        assert subprop.type.stortype is not None

                        subnorm, subinfo = subprop.type.norm(subvalu)

                        edits.append((s_layer.EDIT_PROP_SET, (subprop.name, subnorm, None, subprop.type.stortype), ()))

                propform = self.core.model.form(prop.type.name)
                if propform is not None:

                    doedit = True
                    if self.buidprefetch:
                        node = await self.getNodeByBuid(s_common.buid((propform.name, propnorm)))
                        doedit = node is None

                    if doedit:
                        subedits.extend([x async for x in _getadds(propform, {}, propnorm, typeinfo)])

                edit: s_layer.EditT = (s_layer.EDIT_PROP_SET, (propname, propnorm, None, prop.type.stortype), subedits)
                if propname in formsubs:
                    topsubedits.append(edit)
                else:
                    edits.append(edit)

            buid = s_common.buid((f.name, formnorm))

            if doaddnode:
                # Make all the sub edits for the primary property a conditional nodeEdit under a top-level NODE_ADD
                # edit
                if topsubedits:
                    subnodeedits: s_layer.NodeEditsT = [(buid, f.name, topsubedits)]
                else:
                    subnodeedits = ()
                topedit: s_layer.EditT = (s_layer.EDIT_NODE_ADD, (formnorm, f.type.stortype), subnodeedits)
                yield (buid, f.name, [topedit] + edits)
            else:
                yield (buid, f.name, edits)

        if props is None:
            props = {}

        norm, info = form.type.norm(valu)
        return [x async for x in _getadds(form, props, norm, info, doaddnode=addnode)]

    async def applyNodeEdit(self, edit):
        nodes = await self.applyNodeEdits((edit,))
        return nodes[0]

    async def applyNodeEdits(self, edits):
        '''
        Sends edits to the write layer and evaluates the consequences (triggers, node object updates)
        '''
        if self.readonly:
            mesg = 'The snapshot is in read-only mode.'
            raise s_exc.IsReadOnly(mesg=mesg)

        meta = await self.getSnapMeta()

        todo = s_common.todo('storNodeEdits', edits, meta)
        sodes = await self.core.dyncall(self.wlyr.iden, todo)

        wlyr = self.wlyr
        nodes = []
        callbacks = []
        actualedits = []  # List[Tuple[buid, form, changes]]

        # make a pass through the returned edits, apply the changes to our Nodes()
        # and collect up all the callbacks to fire at once at the end.  It is
        # critical to fire all callbacks after applying all Node() changes.

        for sode in sodes:

            cache = {wlyr.iden: sode}

            node = await self._joinStorNode(sode[0], cache)

            if node is None:
                # We got part of a node but no ndef
                continue

            nodes.append(node)

            postedits = sode[1].get('edits', ())

            if postedits:
                actualedits.append((sode[0], sode[1]['form'], postedits))

            for edit in postedits:

                etyp, parms, _ = edit

                if etyp == s_layer.EDIT_NODE_ADD:
                    node.bylayer['ndef'] = wlyr
                    callbacks.append((node.form.wasAdded, (node,), {}))
                    callbacks.append((self.view.runNodeAdd, (node,), {}))
                    continue

                if etyp == s_layer.EDIT_NODE_DEL:
                    callbacks.append((node.form.wasDeleted, (node,), {}))
                    callbacks.append((self.view.runNodeDel, (node,), {}))
                    continue

                if etyp == s_layer.EDIT_PROP_SET:

                    (name, valu, oldv, stype) = parms

                    prop = node.form.props.get(name)
                    if prop is None: # pragma: no cover
                        logger.warning(f'applyNodeEdits got EDIT_PROP_SET for bad prop {name} on form {node.form}')
                        continue

                    node.props[name] = valu
                    node.bylayer['props'][name] = wlyr

                    callbacks.append((prop.wasSet, (node, oldv), {}))
                    callbacks.append((self.view.runPropSet, (node, prop, oldv), {}))
                    continue

                if etyp == s_layer.EDIT_PROP_DEL:

                    (name, oldv, stype) = parms

                    prop = node.form.props.get(name)
                    if prop is None: # pragma: no cover
                        logger.warning(f'applyNodeEdits got EDIT_PROP_DEL for bad prop {name} on form {node.form}')
                        continue

                    node.props.pop(name, None)
                    node.bylayer['props'].pop(name, None)

                    callbacks.append((prop.wasDel, (node, oldv), {}))
                    callbacks.append((self.view.runPropSet, (node, prop, oldv), {}))
                    continue

                if etyp == s_layer.EDIT_TAG_SET:

                    (tag, valu, oldv) = parms

                    node.tags[tag] = valu
                    node.bylayer['tags'][tag] = wlyr

                    callbacks.append((self.view.runTagAdd, (node, tag, valu), {}))
                    callbacks.append((self.wlyr.fire, ('tag:add', ), {'tag': tag, 'node': node.iden()}))
                    continue

                if etyp == s_layer.EDIT_TAG_DEL:

                    (tag, oldv) = parms

                    node.tags.pop(tag, None)
                    node.bylayer['tags'].pop(tag, None)

                    callbacks.append((self.view.runTagDel, (node, tag, oldv), {}))
                    callbacks.append((self.wlyr.fire, ('tag:del', ), {'tag': tag, 'node': node.iden()}))
                    continue

                if etyp == s_layer.EDIT_TAGPROP_SET:
                    (tag, prop, valu, oldv, stype) = parms
                    node.tagprops[(tag, prop)] = valu
                    node.bylayer['tags'][(tag, prop)] = wlyr
                    continue

                if etyp == s_layer.EDIT_TAGPROP_DEL:
                    (tag, prop, oldv, stype) = parms
                    node.tagprops.pop((tag, prop), None)
                    node.bylayer['tags'].pop((tag, prop), None)
                    continue

        [await func(*args, **kwargs) for (func, args, kwargs) in callbacks]

        if actualedits:
            providen, provstack = self.core.provstor.stor()
            if providen is not None:
                await self.fire('prov:new', time=meta['time'], user=meta['user'], prov=providen, provstack=provstack)
            await self.fire('node:edits', edits=actualedits)

        return nodes

    async def addNode(self, name, valu, props=None):
        '''
        Add a node by form name and value with optional props.

        Args:
            name (str): The form of node to add.
            valu (obj): The value for the node.
            props (dict): Optional secondary properties for the node.

        Notes:
            If a props dictionary is provided, it may be mutated during node construction.

        Returns:
            s_node.Node: A Node object. It may return None if the snap is unable to add or lift the node.
        '''
        if self.readonly:
            mesg = 'The snapshot is in read-only mode.'
            raise s_exc.IsReadOnly(mesg=mesg)

        form = self.core.model.form(name)
        if form is None:
            raise s_exc.NoSuchForm(name=name)

        if form.isrunt:
            raise s_exc.IsRuntForm(mesg='Cannot make runt nodes.',
                                   form=form.full, prop=valu)

        try:

            if self.buidprefetch:
                norm, info = form.type.norm(valu)
                node = await self.getNodeByBuid(s_common.buid((form.name, norm)))
                if node is not None:
                    # TODO implement node.setNodeProps()
                    if props is not None:
                        for p, v in props.items():
                            await node.set(p, v)
                    return node

            adds = await self.getNodeAdds(form, valu, props=props)

        except asyncio.CancelledError: # pragma: no cover
            raise

        except Exception as e:
            if not self.strict:
                await self.warn(f'addNode: {e}')
                return None
            raise

        nodes = await self.applyNodeEdits(adds)

        # Adds is top-down, so the first node is what we want
        return nodes[0]

    async def addFeedNodes(self, name, items):
        '''
        Call a feed function and return what it returns (typically yields Node()s).

        Args:
            name (str): The name of the feed record type.
            items (list): A list of records of the given feed type.

        Returns:
            (object): The return value from the feed function. Typically Node() generator.

        '''
        func = self.core.getFeedFunc(name)
        if func is None:
            raise s_exc.NoSuchName(name=name)

        logger.info(f'adding feed nodes ({name}): {len(items)}')

        genr = func(self, items)
        if not isinstance(genr, types.AsyncGeneratorType):
            if isinstance(genr, types.CoroutineType):
                genr.close()
            mesg = f'feed func returned a {type(genr)}, not an async generator.'
            raise s_exc.BadCtorType(mesg=mesg, name=name)

        async for node in genr:
            yield node

    async def addFeedData(self, name, items):

        func = self.core.getFeedFunc(name)
        if func is None:
            raise s_exc.NoSuchName(name=name)

        logger.info(f'adding feed data ({name}): {len(items)}')

        retn = func(self, items)

        # If the feed function is an async generator, run it...
        if isinstance(retn, types.AsyncGeneratorType):
            retn = [x async for x in retn]
        elif s_coro.iscoro(retn):
            await retn

    async def addTagNode(self, name):
        '''
        Ensure that the given syn:tag node exists.
        '''
        return await self.tagcache.aget(name)

    async def _addTagNode(self, name):
        return await self.addNode('syn:tag', name)

    async def _raiseOnStrict(self, ctor, mesg, **info):
        if self.strict:
            raise ctor(mesg=mesg, **info)
        return False

    async def addNodes(self, nodedefs):
        '''
        Add/merge nodes in bulk.

        The addNodes API is designed for bulk adds which will
        also set properties and add tags to existing nodes.
        Nodes are specified as a list of the following tuples:

            ( (form, valu), {'props':{}, 'tags':{}})

        Args:
            nodedefs (list): A list of nodedef tuples.

        Returns:
            (list): A list of xact messages.
        '''
        for (formname, formvalu), forminfo in nodedefs:
            try:
                props = forminfo.get('props')

                # remove any universal created props...
                if props is not None:
                    props.pop('.created', None)

                oldstrict = self.strict
                self.strict = True
                try:
                    node = await self.addNode(formname, formvalu, props=props)
                    if node is not None:
                        tags = forminfo.get('tags')
                        if tags is not None:
                            for tag, asof in tags.items():
                                await node.addTag(tag, valu=asof)

                        tagprops = forminfo.get('tagprops', {})
                        if tagprops is not None:
                            for tag, props in tagprops.items():
                                for prop, valu in props.items():
                                    try:
                                        await node.setTagProp(tag, prop, valu)
                                    except s_exc.NoSuchTagProp:
                                        mesg = \
                                            f'Tagprop [{prop}] does not exist, cannot set it on [{formname}={formvalu}]'
                                        logger.warning(mesg)
                                        continue

                except Exception as e:
                    if not oldstrict:
                        await self.warn(f'addNodes failed on {formname}, {formvalu}, {forminfo}: {e}')
                        continue
                    raise

                finally:
                    self.strict = oldstrict

                yield node

            except asyncio.CancelledError:  # pragma: no cover
                raise

            except Exception:
                logger.exception(f'Error making node: [{formname}={formvalu}]')

    async def getRuntNodes(self, full, valu=None, cmpr=None):

        todo = s_common.todo('runRuntLift', full, valu, cmpr)
        async for sode in self.core.dyniter('cortex', todo):

            node = s_node.Node(self, sode)
            node.isrunt = True

            yield node

    async def iterNodeEdgesN1(self, buid, verb=None):

        async with await s_spooled.Set.anit(dirn=self.core.dirn) as edgeset:

            for layr in self.layers:

                async for edge in layr.iterNodeEdgesN1(buid, verb=verb):
                    if edge in edgeset:
                        continue

                    await edgeset.add(edge)
                    yield edge

    async def iterNodeEdgesN2(self, buid, verb=None):

        async with await s_spooled.Set.anit(dirn=self.core.dirn) as edgeset:

            for layr in self.layers:

                async for edge in layr.iterNodeEdgesN2(buid, verb=verb):
                    if edge in edgeset:
                        continue

                    await edgeset.add(edge)
                    yield edge

    async def getNodeData(self, buid, name, defv=None):
        '''
        Get nodedata from closest to write layer, no merging involved
        '''
        for layr in reversed(self.layers):
            todo = s_common.todo('getNodeData', buid, name)
            ok, valu = await self.core.dyncall(layr.iden, todo)
            if ok:
                return valu
        return defv

    async def iterNodeData(self, buid):
        '''
        Returns:  Iterable[Tuple[str, Any]]
        '''
        some = False
        for layr in reversed(self.layers):
            todo = s_common.todo('iterNodeData', buid)
            async for item in self.core.dyniter(layr.iden, todo):
                some = True
                yield item
            if some:
                return



import synapse.exc as s_exc
import synapse.tests.utils as s_test

class MacroTest(s_test.SynTest):

    async def test_stormlib_macro(self):

        async with self.getTestCore() as core:

            visi = await core.auth.addUser('visi')
            asvisi = {'user': visi.iden}

            await core.nodes('[ inet:ipv4=1.2.3.4 ]')

            await core.callStorm('macro.set hehe ${ inet:ipv4 }')
            await core.callStorm('macro.set hoho "+#foo"')

            msgs = await core.stormlist('macro.list')

            self.stormIsInPrint('hehe', msgs)
            self.stormIsInPrint('hoho', msgs)
            self.stormIsInPrint('root', msgs)
            self.stormIsInPrint('2 macros found', msgs)

            nodes = await core.nodes('macro.exec hehe', opts=asvisi)
            self.eq(nodes[0].ndef, ('inet:ipv4', 0x01020304))

            nodes = await core.nodes('$name="hehe" | macro.exec $name',)
            self.eq(nodes[0].ndef, ('inet:ipv4', 0x01020304))

            nodes = await core.nodes('macro.exec hehe | macro.exec hoho', opts=asvisi)
            self.len(0, nodes)

            with self.raises(s_exc.StormRuntimeError):
                await core.nodes('[ test:str=hehe ] $name=$node.value() | macro.exec $name')

            with self.raises(s_exc.NoSuchName):
                await core.nodes('macro.exec newp')

            with self.raises(s_exc.NoSuchName):
                await core.nodes('$lib.macro.del(haha)')

            with self.raises(s_exc.AuthDeny):
                await core.nodes('$lib.macro.del(hehe)', opts=asvisi)

            with self.raises(s_exc.AuthDeny):
                await core.nodes('$lib.macro.set(hehe, ${ inet:ipv6 })', opts=asvisi)

            msgs = await core.stormlist('macro.set hehe ${ inet:ipv4 -:asn=30 }')
            self.stormIsInPrint('Set macro: hehe', msgs)

            msgs = await core.stormlist('macro.get hehe')
            self.stormIsInPrint('inet:ipv4 -:asn=30', msgs)

            msgs = await core.stormlist('macro.del hehe')
            self.stormIsInPrint('Removed macro: hehe', msgs)

            self.none(await core.callStorm('return($lib.macro.get(hehe))'))



import synapse.common as s_common

import synapse.tests.utils as s_t_utils

class PsModelTest(s_t_utils.SynTest):
    async def test_ps_simple(self):

        person0 = s_common.guid()
        persona0 = s_common.guid()

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                node = await snap.addNode('ps:tokn', ' BOB ')
                self.eq(node.ndef[1], 'bob')

                node = await snap.addNode('ps:name', ' robert GREY  the\t3rd  ')
                self.eq(node.ndef[1], 'robert grey the 3rd')
                file0 = 'sha256:' + 64 * '0'
                person_props = {
                    'dob': '1971',
                    'img': file0,
                    'nick': 'pennywise',
                    # 'guidname': '', # fixme guid aliases
                    'name': 'robert clown grey',
                    'name:sur': 'grey',
                    'name:middle': 'clown',
                    'name:given': 'robert',
                }
                node = await snap.addNode('ps:person', person0, person_props)
                self.eq(node.ndef[1], person0)
                self.eq(node.get('img'), file0)
                self.eq(node.get('dob'), 31536000000)
                self.eq(node.get('nick'), 'pennywise')
                self.eq(node.get('name'), 'robert clown grey')
                self.eq(node.get('name:sur'), 'grey')
                self.eq(node.get('name:middle'), 'clown')
                self.eq(node.get('name:given'), 'robert')
                # self.eq(node.get('img'), '')  # fixme file:bytes
                # self.eq(node.get('guidname'), '')  # fixme guid aliases

                persona_props = {
                    'dob': '2000',
                    'img': file0,
                    'nick': 'acid burn',
                    'person': person0,
                    'name': 'Эммануэль брат Гольдштейн',
                    'name:sur': 'Гольдштейн',
                    'name:middle': 'брат',
                    'name:given': 'эммануэль',
                }
                node = await snap.addNode('ps:persona', persona0, persona_props)
                self.eq(node.ndef[1], persona0)
                self.eq(node.get('img'), file0)
                self.eq(node.get('dob'), 946684800000)
                self.eq(node.get('nick'), 'acid burn')
                self.eq(node.get('person'), person0)
                self.eq(node.get('name'), 'эммануэль брат гольдштейн')
                self.eq(node.get('name:sur'), 'гольдштейн')
                self.eq(node.get('name:middle'), 'брат')
                self.eq(node.get('name:given'), 'эммануэль')
                # self.eq(node.get('img'), '')  # fixme file:bytes
                # self.eq(node.get('guidname'), '')  # fixme guid aliases

                node = await snap.addNode('ps:person:has', (person0, ('test:str', 'sewer map')))
                self.eq(node.ndef[1], (person0, ('test:str', 'sewer map')))
                self.eq(node.get('person'), person0)
                self.eq(node.get('node'), ('test:str', 'sewer map'))
                self.eq(node.get('node:form'), 'test:str')

                node = await snap.addNode('ps:persona:has', (persona0, ('test:str', 'the gibson')))
                self.eq(node.ndef[1], (persona0, ('test:str', 'the gibson')))
                self.eq(node.get('persona'), persona0)
                self.eq(node.get('node'), ('test:str', 'the gibson'))
                self.eq(node.get('node:form'), 'test:str')

                org0 = s_common.guid()
                con0 = s_common.guid()
                cprops = {
                    'org': org0,
                    'asof': '20080414',
                    'person': person0,
                    'name': 'Tony  Stark',
                    'title': 'CEO',
                    'orgname': 'Stark Industries, INC',
                    # 'img': '',  # fixme file:bytes
                    'user': 'ironman',
                    'web:acct': ('twitter.com', 'ironman'),
                    'dob': '1976-12-17',
                    'url': 'https://starkindustries.com/',
                    'email': 'tony.stark@gmail.com',
                    'email:work': 'tstark@starkindustries.com',
                    'phone': '12345678910',
                    'phone:fax': '12345678910',
                    'phone:work': '12345678910',
                    'address': '1 Iron Suit Drive, San Francisco, CA, 22222, USA',
                }

                node = await snap.addNode('ps:contact', con0, cprops)
                self.eq(node.ndef[1], con0)
                self.eq(node.get('org'), org0)
                self.eq(node.get('asof'), 1208131200000)
                self.eq(node.get('person'), person0)
                self.eq(node.get('name'), 'tony stark')
                self.eq(node.get('title'), 'ceo')
                self.eq(node.get('orgname'), 'stark industries, inc')
                self.eq(node.get('user'), 'ironman')
                self.eq(node.get('web:acct'), ('twitter.com', 'ironman'))
                self.eq(node.get('dob'), 219628800000)
                self.eq(node.get('url'), 'https://starkindustries.com/')
                self.eq(node.get('email'), 'tony.stark@gmail.com')
                self.eq(node.get('email:work'), 'tstark@starkindustries.com')
                self.eq(node.get('phone'), '12345678910')
                self.eq(node.get('phone:fax'), '12345678910')
                self.eq(node.get('phone:work'), '12345678910')
                self.eq(node.get('address'), '1 iron suit drive, san francisco, ca, 22222, usa')





# PE Resource Types
resource_types = (
    (1, 'RT_CURSOR'),
    (2, 'RT_BITMAP'),
    (3, 'RT_ICON'),
    (4, 'RT_MENU'),
    (5, 'RT_DIALOG'),
    (6, 'RT_STRING'),
    (7, 'RT_FONTDIR'),
    (8, 'RT_FONT'),
    (9, 'RT_ACCELERATOR'),
    (10, 'RT_RCDATA'),
    (11, 'RT_MESSAGETABLE'),
    (12, 'RT_GROUP_CURSOR'),
    (14, 'RT_GROUP_ICON'),
    (16, 'RT_VERSION'),
    (17, 'RT_DLGINCLUDE'),
    (19, 'RT_PLUGPLAY'),
    (20, 'RT_VXD'),
    (21, 'RT_ANICURSOR'),
    (22, 'RT_ANIICON'),
    (23, 'RT_HTML'),
    (24, 'RT_MANIFEST'),
)

def getRsrcTypes():
    return resource_types

# Language identifier constants and strings
# https://docs.microsoft.com/en-us/windows/desktop/intl/language-identifier-constants-and-strings#language-identifier-notes
langcodes = (
    (0x0C00, 'custom default'), # Default custom locale language-Default custom sublanguage'
    (0x1400, 'ui_custom_default'), # Default custom MUI locale language-Default custom MUI sublanguage
    (0x007f, 'invariant'), # Invariant locale language-invariant sublanguage
    (0x0000, 'neutral'), # Neutral locale language-Neutral sublanguage
    (0x0800, 'sys default'), # System default locale language-System default sublanguage
    (0x1000, 'custom unspecified'), # Unspecified custom locale language-Unspecified custom sublanguage
    (0x0400, 'default'), # User default locale language-User default sublanguage
    (0x0436, 'af-ZA'), # AFRIKAANS_SOUTH_AFRICA
    (0x041C, 'sq-AL'), # ALBANIAN_ALBANIA
    (0x0484, 'gsw-FR'), # ALSATIAN_FRANCE
    (0x045E, 'am-ET'), # AMHARIC_ETHIOPIA
    (0x1401, 'ar-DZ'), # ARABIC_ALGERIA
    (0x3C01, 'ar-BH'), # ARABIC_BAHRAIN
    (0x0C01, 'ar-EG'), # ARABIC_EGYPT
    (0x0801, 'ar-IQ'), # ARABIC_IRAQ
    (0x2C01, 'ar-JO'), # ARABIC_JORDAN
    (0x3401, 'ar-KW'), # ARABIC_KUWAIT
    (0x3001, 'ar-LB'), # ARABIC_LEBANON
    (0x1001, 'ar-LY'), # ARABIC_LIBYA
    (0x1801, 'ar-MA'), # ARABIC_MOROCCO
    (0x2001, 'ar-OM'), # ARABIC_OMAN
    (0x4001, 'ar-QA'), # ARABIC_QATAR
    (0x0401, 'ar-SA'), # ARABIC_SAUDI_ARABIA
    (0x2801, 'ar-SY'), # ARABIC_SYRIA
    (0x1C01, 'ar-TN'), # ARABIC_TUNISIA
    (0x3801, 'ar-AE'), # ARABIC_UAE
    (0x2401, 'ar-YE'), # ARABIC_YEMEN
    (0x042B, 'hy-AM'), # ARMENIAN_ARMENIA
    (0x044D, 'as-IN'), # ASSAMESE_INDIA
    (0x082C, 'az-AZ-Cyrillic'), # AZERI_CYRILLIC
    (0x042C, 'az-AZ-Latin'), # AZERI_LATIN
    (0x0445, 'bn-Bangledesh'), # BANGLA_BANGLADESH AND BANGLA_INDIA
    (0x0845, 'bn-IN'), # BANGLA_INDIA
    (0x046D, 'ba-RU'), # BASHKIR_RUSSIA
    (0x042D, 'Basque-Basque'), # BASQUE_BASQUE
    (0x0423, 'be-BY'), # BELARUSIAN_BELARUS
    (0x781A, 'bs-neutral'), # BOSNIAN NEUTRAL
    (0x201A, 'bs-BA-Cyrillic'), # BOSNIAN_BOSNIA_HERZEGOVINA_CYRILLIC
    (0x141A, 'bs-BA-Latin'), # BOSNIAN_BOSNIA_HERZEGOVINA_LATIN
    (0x047E, 'br-FR'), # BRETON_FRANCE
    (0x0402, 'bg-BG'), # BULGARIAN_BULGARIA
    (0x0492, 'ku-IQ'), # CENTRAL_KURDISH_IRAQ
    (0x045C, 'chr-Cher'), # CHEROKEE_CHEROKEE
    (0x0403, 'ca-ES'), # CATALAN_CATALAN
    (0x0C04, 'zh-HK'), # CHINESE_HONGKONG
    (0x1404, 'zh-MO'), # CHINESE_MACAU
    (0x1004, 'zh-SG'), # CHINESE_SINGAPORE
    (0x0004, 'zh-Hans'), # CHINESE_SIMPLIFIED
    (0x7C04, 'zh-Hant'), # CHINESE_TRADITIONAL
    (0x0483, 'co-FR'), # CORSICAN_FRANCE
    (0x001A, 'hr'), # CROATIAN Neutral
    (0x101A, 'hr-BA'), # CROATIAN_BOSNIA_HERZEGOVINA_LATIN
    (0x041A, 'hr-HR'), # CROATIAN_CROATIA
    (0x0405, 'cs-CZ'), # CZECH_CZECH_REPUBLIC
    (0x0406, 'da-DK'), # DANISH_DENMARK
    (0x048C, 'prs-AF'), # DARI_AFGHANISTAN
    (0x0465, 'dv-MV'), # DIVEHI_MALDIVES
    (0x0813, 'nl-BE'), # DUTCH_BELGIAN
    (0x0413, 'nl-NL'), # DUTCH DUTCH
    (0x0C09, 'en-AU'), # ENGLISH_AUS
    (0x2809, 'en-BZ'), # ENGLISH_BELIZE
    (0x1009, 'en-CA'), # ENGLISH_CAN
    (0x2409, 'en-029'), # ENGLISH_CARIBBEAN
    (0x4009, 'en-IN'), # ENGLISH_INDIA
    (0x1809, 'en-IE'), # ENGLISH_IRELAND
    (0x2009, 'en-JM'), # ENGLISH_JAMAICA
    (0x4409, 'en-MY'), # ENGLISH_MALAYSIA
    (0x1409, 'en-NZ'), # ENGLISH_NZ
    (0x3409, 'en-PH'), # ENGLISH_PHILIPPINES
    (0x4809, 'en-SG'), # ENGLISH_SINGAPORE
    (0x1c09, 'en-ZA'), # ENGLISH_SOUTH_AFRICA
    (0x2C09, 'en-TT'), # ENGLISH_TRINIDAD
    (0x0809, 'en-GB'), # ENGLISH_UK
    (0x0409, 'en-US'), # ENGLISH_US
    (0x3009, 'en-ZW'), # ENGLISH_ZIMBABWE
    (0x0425, 'et-EE'), # ESTONIAN_ESTONIA
    (0x0438, 'fo-FO'), # FAEROESE_FAROE_ISLANDS
    (0x0464, 'fil-PH'), # FILIPINO_PHILIPPINES
    (0x040B, 'fi-FI'), # FINNISH_FINLAND
    (0x080c, 'fr-BE'), # FRENCH_BELGIAN
    (0x0C0C, 'fr-CA'), # FRENCH_CANADIAN
    (0x040c, 'fr-FR'), # FRENCH_FRENCH
    (0x140C, 'fr-LU'), # FRENCH_LUXEMBOURG
    (0x180C, 'fr-MC'), # FRENCH_MONACO
    (0x100C, 'fr-CH'), # FRENCH_SWISS
    (0x0462, 'fy-NL'), # FRISIAN_NETHERLANDS
    (0x0456, 'gl-ES'), # GALICIAN_GALICIAN
    (0x0437, 'ka-GE'), # GEORGIAN_GEORGIA
    (0x0C07, 'de-AT'), # GERMAN_AUSTRIAN
    (0x0407, 'de-DE'), # GERMAN_GERMAN
    (0x1407, 'de-LI'), # GERMAN_LIECHTENSTEIN
    (0x1007, 'de-LU'), # GERMAN_LUXEMBOURG
    (0x0807, 'de-CH'), # GERMAN_SWISS
    (0x0408, 'el-GR'), # GREEK_GREECE
    (0x046F, 'kl-GL'), # GREENLANDIC_GREENLAND
    (0x0447, 'gu-IN'), # GUJARATI_INDIA
    (0x0468, 'ha-NG'), # HAUSA_NIGERIA_LATIN
    (0x0475, 'haw-US'), # HAWAIIAN_US
    (0x040D, 'he-IL'), # HEBREW_ISRAEL
    (0x0439, 'hi-IN'), # HINDI_INDIA
    (0x040E, 'hu-HU'), # HUNGARIAN_HUNGARY
    (0x040F, 'is-IS'), # ICELANDIC_ICELAND
    (0x0470, 'ig-NG'), # IGBO_NIGERIA
    (0x0421, 'id-ID'), # INDONESIAN_INDONESIA
    (0x085D, 'iu-CA-Latin'), # INUKTITUT_CANADA_LATIN
    (0x045D, 'iu-CA'), # INUKTITUT_CANADA
    (0x083C, 'ga-IE'), # IRISH_IRELAND
    (0x0434, 'xh-ZA'), # XHOSA_SOUTH_AFRICA
    (0x0435, 'zu-ZA'), # ZULU_SOUTH_AFRICA
    (0x0410, 'it-IT'), # ITALIAN_ITALIAN
    (0x0810, 'it-CH'), # ITALIAN_SWISS
    (0x0411, 'ja-JP'), # JAPANESE_JAPAN
    (0x044B, 'kn-IN'), # KANNADA_INDIA
    (0x043F, 'kk-KZ'), # KAZAK_KAZAKHSTAN
    (0x0453, 'kh-KH'), # KHMER_CAMBODIA
    (0x0486, 'qut-GT'), # KICHE_GUATEMALA
    (0x0487, 'rw-RW'), # KINYARWANDA_RWANDA
    (0x0457, 'kok-IN'), # KONKANI_INDIA
    (0x0412, 'ko-KR'), # KOREA_KOREAN
    (0x0440, 'ky-KG'), # KYRGYZ_KYRGYZSTAN
    (0x0454, 'lo-LA'), # LAO_LAO
    (0x0426, 'lv-LV'), # LATVIAN_LATVIA
    (0x0427, 'lt-LT'), # LITHUANIAN_LITHUANIA
    (0x082E, 'dsb-DE'), # LOWER_SORBIAN_GERMANY
    (0x046E, 'lb-LU'), # LUXEMBOURGISH_LUXEMBOURG
    (0x042F, 'mk-MK'), # MACEDONIAN_MACEDONIA
    (0x083E, 'ms-BN'), # MALAY_BRUNEI_DARUSSALAM
    (0x043e, 'ms-MY'), # MALAY_MALAYSIA
    (0x044C, 'ml-IN'), # MALAYALAM_INDIA
    (0x043A, 'mt-MT'), # MALTESE_MALTA
    (0x0481, 'mi-NZ'), # MAORI_NEW_ZEALAND
    (0x047A, 'arn-CL'), # MAPUDUNGUN_CHILE'
    (0x044E, 'mr-IN'), # MARATHI_INDIA
    (0x047C, 'moh-CA'), # MOHAWK_MOHAWK
    (0x0450, 'mn-MN-Cyrllic'), # MONGOLIAN_CYRILLIC_MONGOLIA
    (0x0850, 'mn-MN-Prc'), # MONGOLIAN_PRC
    (0x0461, 'ne-NP'), # NEPALI_NEPAL
    (0x0414, 'nb-NO'), # NORWEGIAN_BOKMAL
    (0x0814, 'no-NO'), # NORWEGIAN_NYNORSK
    (0x0482, 'oc-FR'), # OCCITAN_FRANCE
    (0x0448, 'or-IN'), # ORIYA_INDIA
    (0x0463, 'ps-AF'), # PASHTO_AFGHANISTAN
    (0x0429, 'fa-IR'), # PERSIAN_IRAN
    (0x0415, 'pl-PL'), # POLISH_POLAND
    (0x0416, 'pt-BR'), # PORTUGUESE_BRAZILIAN
    (0x0816, 'pt-PT'), # PORTUGUESE-PORTUGUESE
    (0x0867, 'ff-SN'), # PULAR_SENEGAL
    (0x0446, 'pa-IN'), # PUNJABI_INDIA
    (0x0846, 'pa-PK'), # PUNJABI_PAKISTAN
    (0x046B, 'quz-BO'), # QUECHUA_BOLIVIA
    (0x086B, 'quz-EC'), # QUECHUA_ECUADOR
    (0x0C6B, 'quz-PE'), # QUECHUA_PERU
    (0x0418, 'ro-RO'), # ROMANIAN_ROMANIA
    (0x0417, 'rm-CH'), # ROMANSH_SWITZERLAND
    (0x0419, 'ru-RU'), # RUSSIAN_RUSSIA
    (0x0485, 'sah-RU'), # SAKHA_RUSSIA
    (0x243B, 'smn-FIl'), # SAMI_INARI_FINLAND
    (0x103B, 'smj-NO'), # SAMI_LULE_NORWAY
    (0x143B, 'smj-SE'), # SAMI_LULE_SWEDEN
    (0x0C3B, 'se-FI'), # SAMI_NORTHERN_FINLAND
    (0x043B, 'se-NO'), # SAMI_NORTHERN_NORWAY
    (0x083B, 'se-SE'), # SAMI_NORTHERN_SWEDEN
    (0x203B, 'sms-FI'), # SAMI_SKOLT_FINLAND
    (0x183B, 'sma-NO'), # SAMI_SOUTHERN_NORWAY
    (0x1C3B, 'sma-SE'), # SAMI_SOUTHERN_SWEDEN
    (0x044F, 'sa-IN'), # SANSKRIT_INDIA
    (0x7C1A, 'sr-Neutral'), # SERBIAN_NEUTRAL
    (0x1C1A, 'sr-BA'), # SERBIAN_BOSNIA_HERZEGOVINA_CYRILLIC
    (0x181A, 'sr-code-Latin'), # SERBIAN_BOSNIA_HERZEGOVINA_LATIN
    (0x0C1A, 'sr-CS-Cyrillic'), # SERBIAN_CYRILLIC
    (0x081A, 'sr-CS-Latin'), # SERBIAN_LATIN
    (0x046C, 'nso-ZA'), # SOTHO_NORTHERN_SOUTH_AFRICA
    (0x0832, 'tn-BW'), # TSWANA_BOTSWANA
    (0x0432, 'tn-ZA'), # TSWANA_SOUTH_AFRICA
    (0x0859, 'sd-PK'), # SINDHI_PAKISTAN
    (0x045B, 'si-LK'), # SINHALESE_SRI_LANKA
    (0x041B, 'sk-SK'), # SLOVAK_SLOVAKIA
    (0x0424, 'sl-SI'), # SLOVENIAN_SLOVENIA
    (0x2C0A, 'es-AR'), # SPANISH_ARGENTINA
    (0x400A, 'es-BO'), # SPANISH_BOLIVIA
    (0x340A, 'es-CL'), # SPANISH_CHILE
    (0x240A, 'es-CO'), # SPANISH_COLOMBIA
    (0x140A, 'es-CR'), # SPANISH_COSTA_RICA
    (0x1C0A, 'es-DO'), # SPANISH_DOMINICAN_REPUBLIC
    (0x300A, 'es-EC'), # SPANISH_ECUADOR
    (0x440A, 'es-SV'), # SPANISH_EL_SALVADOR
    (0x100A, 'es-GT'), # SPANISH_GUATEMALA
    (0x480A, 'es-HN'), # SPANISH_HONDURAS
    (0x080A, 'es-MX'), # SPANISH_MEXICAN
    (0x4C0A, 'es-NI'), # SPANISH_NICARAGUA
    (0x180A, 'es-PA'), # SPANISH_PANAMA
    (0x3C0A, 'es-PY'), # SPANISH_PARAGUAY
    (0x280A, 'es-PE'), # SPANISH_PERU
    (0x500A, 'es-PR'), # SPANISH_PUERTO_RICO
    (0x0C0A, 'es-ES-modern'), # SPANISH_MODERN
    (0x040A, 'es-ES-traditional'), # SPANISH
    (0x540A, 'es-US'), # SPANISH_US
    (0x380A, 'es-UY'), # SPANISH_URUGUAY
    (0x200A, 'es-VE'), # SPANISH_VENEZUELA
    (0x0441, 'sw-KE'), # SWAHILI
    (0x081D, 'sv-FI'), # SWEDISH_FINLAND
    (0x041D, 'sv-SE'), # SWEDISH_SWEDEN
    (0x045A, 'syr-SY'), # SYRIAC
    (0x0428, 'tg-TJ'), # TAJIK_TAJIKISTAN
    (0x085F, 'tzm-DZ'), # TAMAZIGHT_ALGERIA_LATIN
    (0x0449, 'ta-IN'), # TAMIL_INDIA
    (0x0849, 'ta-LK'), # TAMIL_SRI_LANKA
    (0x0444, 'tt-RU'), # TATAR_RUSSIA
    (0x044A, 'te-IN'), # TELUGU_INDIA
    (0x041E, 'th-TH'), # THAI_THAILAND
    (0x0451, 'bo-CN'), # TIBETAN_PRC
    (0x0873, 'ti-ER'), # TIGRINYA_ERITREA
    (0x0473, 'ti-ET'), # TIGRINYA_ETHIOPIA
    (0x041F, 'tr-TR'), # TURKISH_TURKEY
    (0x0442, 'tk-TM'), # TURKMEN_TURKMENISTAN
    (0x0422, 'uk-UA'), # UKRAINIAN_UKRAINE
    (0x042E, 'hsb-DE'), # UPPER_SORBIAN_GERMANY
    (0x0820, 'ur-IN'), # URDU_INDIA
    (0x0420, 'ur-PK'), # URDU_PAKISTAN
    (0x0480, 'ug-CN'), # UIGHUR_PRC
    (0x0843, 'uz-UZ-Cyrillic'), # UZBEK_CYRILLIC
    (0x0443, 'uz-UZ-Latin'), # UZBEK_LATIN
    (0x0803, 'ca-ES-Valencia'), # VALENCIAN_VALENCIA
    (0x042A, 'vi-VN'), # VIETNAMESE_VIETNAM
    (0x0452, 'cy-GB'), # WELSH_UNITED_KINGDOM
    (0x0488, 'wo-SN'), # WOLOF_SENEGAL
    (0x0478, 'ii-CN'), # YI_PRC
    (0x046A, 'yo-NG'), # YORUBA_NIGERIA
)

def getLangCodes():
    return langcodes
